import sys
import os
this_dir = os.path.dirname(os.path.abspath(__file__))
lib_dir = os.path.abspath(this_dir + '/..')
print("lib_dir=", lib_dir)
sys.path.append(lib_dir)
from Tutorials.BH.BasinHopping import basinhopping
from Tutorials.BH.Customize import CustomTakeStep

# basin hopping global optimization for the ackley multimodal objective function
# from scipy.optimize import basinhopping
import numpy as np
from numpy.random import rand
from numpy import exp
from numpy import sqrt
from numpy import cos
from numpy import e
from numpy import pi
from numpy import meshgrid
from numpy import arange
import matplotlib.pyplot as plt
from matplotlib import animation

# objective function
def objective(v):
    x, y = v
    return -20.0 * exp(-0.2 * sqrt(0.5 * (x**2 + y**2))) - exp(0.5 * (cos(2 * pi * x) + cos(2 * pi * y))) + e + 20

# define range for input
r_min, r_max = -5.0, 5.0
# define the starting point as a random sample from the domain

minima_list = []
init_points = [(-4, -4), (-4, -4), (4, -4), (4, -4), (4, 4), (4, 4), (-4, 4), (-4, 4)]
seed_list = [819526, 497324, 586929, 840661, 807396, 266287, 851844, 274206]

def solve(k):
    # pt = r_min + rand(2) * (r_max - r_min)
    pt = init_points[k]
    fv = objective(pt)
    minima_list.append((pt, fv, True))

    if True:
        def minimum_callback(x, fv, accept):
            # print(x, fv, accept)
            minima_list.append((x, fv, accept))

        # perform the basin hopping search
        # seed = np.random.randint(100000, 999999)
        seed = seed_list[k]
        print("seed=", seed)
        take_step = CustomTakeStep(stepsize=0.5, seed=seed)
        rng = take_step.get_rng()
        result = basinhopping(objective, pt, take_step=take_step, niter=18, callback=minimum_callback, rng=rng)
    else:
        from scipy.optimize import differential_evolution
        def minimum_callback(xk, convergence=0):
            # print(x, fv, accept)
            fv = objective(xk)
            minima_list.append((xk, fv, convergence))

        # define range for input
        r_min, r_max = -5.0, 5.0
        # define the bounds on the search
        bounds = [[r_min, r_max], [r_min, r_max]]
        # perform the differential evolution search
        result = differential_evolution(objective, bounds, callback=minimum_callback)
    return result

for k in range(len(init_points)):
    result = solve(k)
    # summarize the result
    print('Status : %s' % result['message'])
    print('Total Evaluations: %d' % result['nfev'])
    # evaluate solution
    solution = result['x']
    evaluation = objective(solution)
    print('Solution: f(%s) = %.5f' % (solution, evaluation))

# sample input range uniformly at 0.1 increments
xaxis = arange(r_min, r_max, 0.1)
yaxis = arange(r_min, r_max, 0.1)
# create a mesh from the axis
xx, yy = meshgrid(xaxis, yaxis)
# compute targets
results = objective((xx, yy)) 
# create a surface plot with the jet color scheme
fig = plt.figure(figsize=(16,8))
ax = fig.add_subplot(121, projection='3d')
ax2 = fig.add_subplot(122)

fig.suptitle("Demonstration of Basin Hopping Algorithm with Ackley Function", fontsize=20)
ax.set_title("Surface Plot", fontsize=16)
ax2.set_title("Contour Plot", fontsize=16)
fig.tight_layout()

surface = ax.plot_surface(xx, yy, results, cmap='jet', alpha=0.3)
contuor = ax2.contourf(xx, yy, results, cmap='jet', alpha=0.3)
# show the plot

minima_array = np.array([np.concatenate([x, [v]]) for x, v, a in minima_list])
x, y, z = minima_array[0]
points, = ax.plot([x], [y], [z], 'o', color='cyan')
trace3, = ax.plot(*minima_array[0:1,:].T, color='red')
point, = ax2.plot(x, y, 'o', color='cyan')
trace2, = ax2.plot(*minima_array[0:1,0:2].T, color='red')

print("len(minima_list)=", len(minima_list))

def reset():
    x, y, z = minima_array[0]
    points.set_data(np.array([x]), np.array([y]))
    points.set_3d_properties(np.array([z]), 'z')
    trace3.set_data(*minima_array[0:1,0:2].T)
    trace3.set_3d_properties(minima_array[0:1,2], 'z')
    point.set_data(x, y)
    trace2.set_data(*minima_array[0:1, 0:2].T)
    return points, trace3, point, trace2

def animate(i):
    x, y, z = minima_array[i]
    points.set_data(np.array([x]), np.array([y]))
    points.set_3d_properties(np.array([z]), 'z')
    point.set_data(x, y)
    color = 'cyan' if i % 20 == 0 else 'red'
    points.set_color(color)
    point.set_color(color)
    j, r = divmod(i, 20)
    start = j*20
    stop = start + r + 1
    slice_ = slice(start, stop)
    trace3.set_data(*minima_array[slice_,0:2].T)
    trace3.set_3d_properties(minima_array[slice_,2], 'z')
    trace2.set_data(*minima_array[slice_,0:2].T)
    return points, trace3, point, trace2

anim = animation.FuncAnimation(fig, animate, frames=len(minima_list), blit=True, init_func=reset, interval=200)
# anim.save("anim.gif")

plt.show()




"""
    CommandMainUtils.py
"""
import sys
from subprocess import call
from .AtsasTools import autorg       as autorg_atsas
from .CommandMain import CommandMain
# import SerialAnalyzer
from molass_legacy.Reports.ReportUtils import make_record

sys_argv_init = sys.argv

def do_CommandMain( folder, file, out_file, subprocess=False, atsas=False, fh=None, run_dir=None, bico_mono_ratio=True ):
    path    = folder + '/' + file

    if subprocess:
        exe_name = 'autorg_kek.bat'
        atsas_arg = '-A' if atsas else ''
        ret = call( '%s/%s %s -o %s %s' % ( run_dir, exe_name, atsas_arg, out_file, path ), shell=True )
        # eq_( ret, 0 )
    else:
        sys.argv = sys_argv_init + [ '-o', out_file, path ]
        command = CommandMain()
        result = command.execute( path, out_file, robust=True, optimize=True )

        orig_result, eval_result    = autorg_atsas( path )

        rec = make_record( path, result, orig_result, eval_result )

        fh.write( rec + '\n' )



# coding: utf-8
"""
    MpInverseAnim.py

    Copyright (c) 2020,2024, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.patches import Polygon, Circle
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.TkSupplements import tk_set_icon_portable
from molass_legacy.KekLib.OurMatplotlib import get_color
from OurManim import manim_init, use_default_style, Animation, Collection, TextGroup, Parallelogram, Circles
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar

class MpInverseAnim(Dialog):
    def __init__(self, parent=None):
        if parent is None:
            from molass_legacy.KekLib.TkUtils import get_tk_root
            parent = get_tk_root()
        self.parent = parent
        self.num_frames = 85
        manim_init()
        Dialog.__init__(self, parent, "Matrix Inverse Tutorial", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):
        tk_set_icon_portable(self)

        cframe = Tk.Frame(body_frame)
        cframe.pack()
        gs = GridSpec(7,3)
        self.fig = fig = plt.figure(figsize=(18,7))
        self.ax__ = fig.add_subplot(gs[0,:])
        self.ax__.set_axis_off()
        self.ax__.text(0.5, 0.5, "Matrix Inverse in Linear Transformation Perspective", ha='center', va='center', fontsize=40)

        self.axes = []
        for k in range(3):
            ax = fig.add_subplot(gs[1:,k])
            self.axes.append(ax)
        # self.fig, self.axes = plt.subplots(nrows=1, ncols=3, figsize=(18,6))
        for ax in  self.axes[1:3]:
            ax.set_axis_off()
        self.mpl_canvas = FigureCanvasTkAgg(self.fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)
        self.toolbar = NavigationToolbar(self.mpl_canvas, cframe)
        self.toolbar.update()
        self.draw()
        self.fig.tight_layout()
        self.mpl_canvas.draw()

    def buttonbox(self):
        box = Tk.Frame(self)
        box.pack()

        w = Tk.Button(box, text="OK", width=10, command=self.ok, default=Tk.ACTIVE)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        w = Tk.Button(box, text="Save", width=10, command=self.save)
        w.pack(side=Tk.LEFT, padx=5, pady=5)

    def draw(self):
        ax1, ax2, ax3 = self.axes

        cube_labels = ['(0,0)', '(1,0)', '(1,1)', '(0,1)']
        cube_points = [(0,0), (1,0), (1,1), (0,1)]
        data_labels = ['(a,b)', '(a+c, b+d)', '(c,d)']
        data_points = [(1,1), (2, 0), (1, -1)]
        data_labels2 = ['(2, 0)', '(1,-1)']
        data_points2 = [(2, 0), (1, -1)]

        labels = cube_labels + data_labels + data_labels2
        points = cube_points + data_points + data_points2
        d = 0.05

        for t, p in zip(labels, points):
            ax1.plot(*p,'o')
            if t > '(a':
                va = 'top'
                yd = -0.05
            else:
                va = 'bottom'
                yd = 0
            ax1.text(p[0]+d, p[1]+d+yd, t, fontsize=16, va=va)

        ax1.set_xlim(-1.0, 2.5)
        ax1.set_ylim(-1.5, 2.0)
        ax1.set_aspect('equal')

        cube = Polygon(cube_points, alpha=0.3, color='pink')
        ax1.add_patch(cube)

        matrix0_1 = r"$ v = \begin{bmatrix} a & c \\ b & d \end{bmatrix} u $"
        ax2.text(0.5, 0.8, matrix0_1, fontsize=30, ha='center', va='center', color='w')
        matrix0_2 = r"$ v = \begin{bmatrix} 1 & 1 \\ 1 & -1 \end{bmatrix} u $"
        ax3.text(0.5, 0.8, matrix0_2, fontsize=30, ha='center', va='center', color='w')

        matrix1_11 = r"$ \begin{bmatrix} \_ \\ \_ \end{bmatrix} = \begin{bmatrix} a & c \\ b & d \end{bmatrix} \begin{bmatrix} 1 \\ 0 \end{bmatrix} $"
        matrix1_12 = r"$ \begin{bmatrix} \_ \\ \_ \end{bmatrix} = \begin{bmatrix} 1 & 1 \\ 1 & -1 \end{bmatrix} \begin{bmatrix} 1 \\ 0 \end{bmatrix} $"
        matrix1_21 = r"$ \begin{bmatrix} a \\ \_ \end{bmatrix} = \begin{bmatrix} a & c \\ b & d \end{bmatrix} \begin{bmatrix} 1 \\ 0 \end{bmatrix} $"
        matrix1_22 = r"$ \begin{bmatrix} 1 \\ \_ \end{bmatrix} = \begin{bmatrix} 1 & 1 \\ 1 & -1 \end{bmatrix} \begin{bmatrix} 1 \\ 0 \end{bmatrix} $"
        matrix1_31 = r"$ \begin{bmatrix} a \\ b \end{bmatrix} = \begin{bmatrix} a & c \\ b & d \end{bmatrix} \begin{bmatrix} 1 \\ 0 \end{bmatrix} $"
        matrix1_32 = r"$ \begin{bmatrix} 1 \\ 1 \end{bmatrix} = \begin{bmatrix} 1 & 1 \\ 1 & -1 \end{bmatrix} \begin{bmatrix} 1 \\ 0 \end{bmatrix} $"

        mt11 = ax2.text(0.5, 0.5, matrix1_11, fontsize=30, ha='center', va='center', color='w')
        mt12 = ax3.text(0.5, 0.5, matrix1_12, fontsize=30, ha='center', va='center', color='w')

        matrix2_11 = r"$ \begin{bmatrix} \_ \\ \_ \end{bmatrix} = \begin{bmatrix} a & c \\ b & d \end{bmatrix} \begin{bmatrix} 0 \\ 1 \end{bmatrix} $"
        matrix2_12 = r"$ \begin{bmatrix} \_ \\ \_ \end{bmatrix} = \begin{bmatrix} 1 & 1 \\ 1 & -1 \end{bmatrix} \begin{bmatrix} 0 \\ 1 \end{bmatrix} $"
        matrix2_21 = r"$ \begin{bmatrix} c \\ \_ \end{bmatrix} = \begin{bmatrix} a & c \\ b & d \end{bmatrix} \begin{bmatrix} 0 \\ 1 \end{bmatrix} $"
        matrix2_22 = r"$ \begin{bmatrix} 1 \\ \_ \end{bmatrix} = \begin{bmatrix} 1 & 1 \\ 1 & -1 \end{bmatrix} \begin{bmatrix} 0 \\ 1 \end{bmatrix} $"
        matrix2_31 = r"$ \begin{bmatrix} c \\ d \end{bmatrix} = \begin{bmatrix} a & c \\ b & d \end{bmatrix} \begin{bmatrix} 0 \\ 1 \end{bmatrix} $"
        matrix2_32 = r"$ \begin{bmatrix} 1 \\ -1 \end{bmatrix} = \begin{bmatrix} 1 & 1 \\ 1 & -1 \end{bmatrix} \begin{bmatrix} 0 \\ 1 \end{bmatrix} $"

        mt21 = ax2.text(0.5, 0.2, matrix2_11, fontsize=30, ha='center', va='center', color='w')
        mt22 = ax3.text(0.5, 0.2, matrix2_12, fontsize=30, ha='center', va='center', color='w')

        def on_reset():
            print('on_reset')
            mt11.set_text(matrix1_11)
            mt12.set_text(matrix1_12)

        collection = Collection(self.num_frames, ret_artists=[mt11, mt12, mt21, mt22], on_reset=on_reset)

        anim1_4 = Animation(0, 20)
        circles1 = Circles(ax1, [cube_points[1]], radius=0.03, color='red')
        circles1.set_target_positions([data_points[0]])
        anim1_4.append(circles1)
        collection.append(anim1_4)

        # a <= (1,0)
        v10 = (['$1$', '$0$'], [(0.67, 0.53), (0.67, 0.44)])
        tg1 = TextGroup(ax2, *v10, fontsize=30)
        tg1.set_target_positions([(0.47, 0.53), (0.54, 0.53)])
        tg1_ = TextGroup(ax3, *v10, fontsize=30)
        tg1_.set_target_positions([(0.47, 0.53), (0.54, 0.53)])

        anim1 = Animation( 0,  5)
        anim1.append(tg1)
        anim1.append(tg1_)
        collection.append(anim1)

        # a <= a
        tg2 = TextGroup(ax2, ['$a$'], [(0.47, 0.53)], fontsize=30)
        tg2.set_target_positions([(0.26, 0.53)])
        tg2_ = TextGroup(ax3, ['$1$'], [(0.47, 0.53)], fontsize=30)
        tg2_.set_target_positions([(0.26, 0.53)])

        def on_stop_anim2():
            mt11.set_text(matrix1_21)
            mt12.set_text(matrix1_22)

        anim2 = Animation( 5, 10, on_stop=on_stop_anim2 )
        anim2.append(tg2)
        anim2.append(tg2_)
        collection.append(anim2)

        # b <= (1,0)
        tg3 = TextGroup(ax2, *v10, fontsize=30)
        tg3.set_target_positions([(0.47, 0.44), (0.54, 0.44)])
        tg3_ = TextGroup(ax3, *v10, fontsize=30)
        tg3_.set_target_positions([(0.47, 0.44), (0.54, 0.44)])

        anim3 = Animation(10, 15)
        anim3.append(tg3)
        anim3.append(tg3_)
        collection.append(anim3)

        # b <= b
        tg4 = TextGroup(ax2, ['$b$'], [(0.47, 0.43)], fontsize=30)
        tg4.set_target_positions([(0.26, 0.43)])
        tg4_ = TextGroup(ax3, ['$1$'], [(0.47, 0.43)], fontsize=30)
        tg4_.set_target_positions([(0.26, 0.43)])

        def on_stop_anim4():
            mt11.set_text(matrix1_31)
            mt12.set_text(matrix1_32)

        anim4 = Animation(15, 20, on_stop=on_stop_anim4 )
        anim4.append(tg4)
        anim4.append(tg4_)
        collection.append(anim4)

        anim5_8 = Animation(20, 40)
        circles2 = Circles(ax1, [cube_points[3]], radius=0.03, color='red')
        circles2.set_target_positions([data_points[2]])
        anim5_8.append(circles2)
        collection.append(anim5_8)

        # c <= (0,1)
        v01 = ['$0$', '$1$'], [(0.67, 0.23), (0.67, 0.13)]
        tg5 = TextGroup(ax2, *v01, fontsize=30)
        tg5.set_target_positions([(0.47, 0.23), (0.54, 0.23)])
        tg5_ = TextGroup(ax3, *v01, fontsize=30)
        tg5_.set_target_positions([(0.47, 0.23), (0.54, 0.23)])
        anim5 = Animation(20, 25)
        anim5.append(tg5)
        anim5.append(tg5_)
        collection.append(anim5)

        # c <= c
        tg6 = TextGroup(ax2, ['$c$'], [(0.54, 0.23)], fontsize=30)
        tg6.set_target_positions([(0.26, 0.23)])
        tg6_ = TextGroup(ax3, ['$1$'], [(0.54, 0.23)], fontsize=30)
        tg6_.set_target_positions([(0.26, 0.23)])

        def on_stop_anim6():
            mt21.set_text(matrix2_21)
            mt22.set_text(matrix2_22)

        anim6 = Animation(25, 30, on_stop=on_stop_anim6 )
        anim6.append(tg6)
        anim6.append(tg6_)
        collection.append(anim6)

        # d <= (0,1)
        tg7 = TextGroup(ax2, *v01, fontsize=30)
        tg7.set_target_positions([(0.47, 0.13), (0.54, 0.13)])
        tg7_ = TextGroup(ax3, *v01, fontsize=30)
        tg7_.set_target_positions([(0.47, 0.13), (0.54, 0.13)])
        anim7 = Animation(30, 35)
        anim7.append(tg7)
        anim7.append(tg7_)
        collection.append(anim7)

        # d <= d
        tg8 = TextGroup(ax2, ['$d$'], [(0.54, 0.13)], fontsize=30)
        tg8.set_target_positions([(0.26, 0.13)])
        tg8_ = TextGroup(ax3, ['$-1$'], [(0.54, 0.13)], fontsize=30)
        tg8_.set_target_positions([(0.26, 0.13)])

        def on_stop_anim8():
            mt21.set_text(matrix2_31)
            mt22.set_text(matrix2_32)

        anim8 = Animation(35, 40, on_stop=on_stop_anim8 )
        anim8.append(tg8)
        anim8.append(tg8_)
        collection.append(anim8)

        anim9 = Animation(40, 45)
        collection.append(anim9)

        # print('index=', collection.index)
        para1 = Parallelogram(ax1, cube_points, alpha=0.3, color='cyan')
        para1.set_target_vertices([(0,0)] + data_points)
        anim_10 = Animation(45, 60)
        anim_10.append(para1)
        collection.append(anim_10)

        anim_11 = Animation(60, 75)
        para1_r = para1.make_reversed(ax1, alpha=0.3, color='cyan')
        para2 = Parallelogram(ax1, cube_points, alpha=0.3, color='yellow')
        inv = np.linalg.inv(np.array([[1,1], [1,-1]]))
        para2.set_target_vertices([(0,0), inv[:,0], inv[:,0]+inv[:,1], inv[:,1]])

        anim_11.append(para1_r)
        anim_11.append(para2)

        collection.append(anim_11)

        self.anim = collection.make_animation(self.fig)

    def save(self):
        import os
        import molass_legacy.KekLib.CustomMessageBox as MessageBox
        path = os.path.join(os.getcwd(), "anim.mp4")
        self.anim.save(path, writer="ffmpeg")
        MessageBox.showinfo("Save Notification", 'The movie has been saved to "%s"' % path, parent=self)



"""
    ErrorCorrection.py

    Copyright (c) 2023-2025, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize
from scipy.interpolate import UnivariateSpline
from molass_legacy.Trimming.Sigmoid import sigmoid
from molass_legacy.DENSS.DenssUtils import fit_data

def compute_corrected_error(qv, i, P, c1, c2, P_, Pe, bq_bounds_, coerced_bq_, debug=False):

    aq, bq = P.T[0:2]
    aq_, bq_ = P_.T[0:2]

    ae = Pe[:,0]
    p0 = np.zeros(Pe.shape[0])
    modified = np.max([p0, (bq - coerced_bq_)*c2/c1], axis=0)
    modified_ratio = modified/ae

    p1 = np.ones(len(aq))
    corrected_ratio = np.min([p1, modified_ratio], axis=0)

    def sigmoid_fit_0(p):
        sy = sigmoid(qv, *p)
        return np.sum((sy - modified_ratio)**2) + np.sum(sy[0:i]**2)*1e5

    ret0 = minimize(sigmoid_fit_0, (1, 0.1, 1, 0))

    def sigmoid_fit_1(p):
        sy = sigmoid(qv, *p)
        return np.sum((sy - corrected_ratio)**2) + np.sum(sy[0:i]**2)*1e5

    ret1 = minimize(sigmoid_fit_1, ret0.x)

    def compute_delta(qv, aq, ae):
        qc, Ic, Icerr, Dc = fit_data(qv, aq, ae)
        spline = UnivariateSpline(qc, Ic, s=0, ext=3)
        delta = (aq - spline(qv))/ae
        return delta

    ae = Pe[:,0]
    delta = compute_delta(qv, aq, ae)

    correct_scale = 2.14        # this scale should be stable for various data

    ae_ = ae * (1 - sigmoid(qv, *ret1.x)*correct_scale)
    # how abount be?

    delta_ = compute_delta(qv, aq_, ae_)

    adjust_scale = np.std(delta) / np.std(delta_)

    if debug:
        import molass_legacy.KekLib.DebugPlot as plt
        print("adjust_scale=", adjust_scale)
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_title("Correction Sigmoid Model Estimation", fontsize=20)
            ax.plot(qv, bq_, color="pink")

            for k, bound in enumerate(bq_bounds_):
                label = "B(q) Bounds" if k == 0 else None
                ax.plot(qv, bound, ":", color="red", label=label)

            ax.set_ylim(-3, 3)

            axt = ax.twinx()
            axt.grid(False)

            axt.plot(qv, modified_ratio, label="A(q) error modified ratio")
            axt.plot(qv, corrected_ratio, label="A(q) error corrected ratio")
            cy = sigmoid(qv, *ret1.x)
            axt.plot(qv, cy, lw=5, label="Correction ratio sigmoid")
            axt.plot(qv, cy*correct_scale, ":", lw=5, label="Effective correction ratio sigmoid (scale=%g)" % correct_scale)

            axt.legend(fontsize=16)

            fig.tight_layout()
            plt.show()

    Pe_ = Pe.copy()
    Pe_[:,0] = ae_ / adjust_scale
    return Pe_, adjust_scale



"""

    ProgressInfoDialog.py

    Copyright (c) 2017-2023, Masatsuyo Takahashi, KEK-PF

"""

import os
import re
import warnings
import time
import logging
from io import StringIO

from molass_legacy.KekLib.BasicUtils             import get_caller_module
from molass_legacy.KekLib.OurTkinter             import Tk, Dialog, ttk, Font, ScrolledText
from molass_legacy.KekLib.TkSupplements          import tk_set_icon_portable, BlinkingFrame, FLASH_STOP_JUST
from molass_legacy.KekLib.TkUtils                import split_geometry, join_geometry
import OurMessageBox        as MessageBox
from Settings               import get_setting
from ProgressInfo           import STATE_INIT, STATE_DOING, STATE_DONE, STATE_ERROR, STATE_FATAL, SAFE_FINISH, cleanup_queues

COLUMN_TITLES   = [ 'stage', 'bar', 'time', 'todo', 'done', 'state' ]
TEXT_TAGS       = { logging.ERROR:'ERROR_TAG', logging.WARNING:'WARNING_TAG' }
TAG_COLORS      = { 'ERROR_TAG':'red', 'WARNING_TAG':'orange' }
STATE_TEXTS     = {
    STATE_INIT  : '',
    STATE_DOING : 'doing',
    STATE_DONE  : 'done',
    STATE_ERROR : 'error',
    STATE_FATAL : 'error',
    }
GREEN   = 'green4'

def make_disp_time( t_ ):
    m_ = t_ // 60
    s_ = t_ % 60
    disp_time = '%2d′%2d″' % ( m_, s_ )
    return disp_time

class ProgressLoghandler( logging.Handler ):
    def __init__( self, log_records ):
        logging.Handler.__init__( self )
        self.log_records  = log_records

    def emit( self, record ):
        if record.levelno >= logging.INFO:
            self.log_records.append( record )

class ProgressInfoDialog( Dialog ):
    def __init__( self, parent, title, pinfo, description_cb=None, stream_labels=None,
                    plot_init_cb=None, plot_update_cb=None, refresh_log_cb=None,
                    plot_final_cb=None,
                    is_alive_cb=None,
                    refresh_interval=100,
                    logger=None ):
        self.grab       = 'local'     # used in grab_set
        self.parent = parent
        self.title_ = title
        self.pinfo  = pinfo
        self.description_cb = description_cb
        self.plot_init_cb   = plot_init_cb
        self.plot_update_cb = plot_update_cb
        self.refresh_log_cb = refresh_log_cb
        self.plot_final_cb  = plot_final_cb
        self.is_alive_cb = is_alive_cb
        self.worker_is_alive = True
        self.stream_labels  = stream_labels
        self.fixed_font = Font.Font( family="Courier", size=9 )
        self.stop_mode  = False
        self.caller_module = get_caller_module( level=2 )
        self.is_at_end_ = False
        self.refresh_interval = refresh_interval
        self.logger = logger
        Dialog.__init__( self, self.parent, self.title_, visible=False )

    def show( self, terminate_cb=None, logger=None ):
        cleanup_queues()        # just in case when there remain previous messages
        self.show_return = None
        self.terminate_cb = terminate_cb
        self.log_records = []
        if logger is not None:
            log_handler = ProgressLoghandler( self.log_records )
            logger.addHandler( log_handler )

        # need to use parent
        # because self is not yet a tk widget before the __init__ call
        self.monitoring_loop = True
        self.can_safely_finish = False
        self.parent.after( self.refresh_interval, self.refresh )
        self.stream_start_time = time.time()
        self._show()

        if logger is not None:
            logger.removeHandler( log_handler )
            log_handler = None

        return self.show_return

    def body( self, body_frame ):
        tk_set_icon_portable( self, module=self.caller_module )

        self.vars_ = []
        self.bars  = []
        self.blinks = []

        iframe = Tk.Frame( body_frame )
        iframe.pack( padx=40 )

        if self.description_cb is not None:
            self.description_cb( self, iframe )

        progress_frame  = Tk.Frame( iframe )
        progress_frame.pack( fill=Tk.X, pady=10 )
        progress_label  = Tk.Label( progress_frame, text="Progress", width=12, anchor=Tk.W )
        progress_label.grid( row=0, column=0, sticky=Tk.N + Tk.W )
        progress_table  = Tk.Frame( progress_frame )
        progress_table.grid( row=0, column=1, sticky=Tk.E )

        for j, col_title in enumerate( COLUMN_TITLES ):
            label   = Tk.Label( progress_table, text=col_title )
            label.grid( row=0, column=j, padx=5 )

        row = 0
        self.first_stream = None
        self.num_buttons = 0
        self.button_states = []
        for i in range( len( self.pinfo.stream_max_dict ) ):
            tim_var = Tk.StringVar()
            max_var = Tk.StringVar()
            val_var = Tk.StringVar()
            stt_var = Tk.StringVar()
            self.vars_.append( [ tim_var, max_var, val_var, stt_var ] )
            self.button_states.append(0)
            self.num_buttons += 1
            stream_size = self.pinfo.stream_max_dict[i]

            bar_label = 'Bar ' + str(i) if self.stream_labels is None else self.stream_labels[i]
            label   = Tk.Label( progress_table, text=bar_label, font=self.fixed_font )
            bar = ttk.Progressbar( progress_table, orient='horizontal', length=200, mode='determinate' )
            self.bars.append( bar )
            state_label = Tk.Label( progress_table, textvariable=stt_var, justify=Tk.LEFT, fg=GREEN )
            objects = []
            objects.append( [ state_label, { 'row':row+1, 'column':5, 'sticky':Tk.W } ] )
                                        # row, column is based on this parent iframe1
            blink = BlinkingFrame( progress_table, objects, grid=True, flash_stop_type=FLASH_STOP_JUST )
            self.blinks.append( [ state_label, blink ] )

            if stream_size == 0:
                # no grid call is done while the objects exist
                continue

            if self.first_stream is None:
                self.first_stream = i
            row += 1
            label.grid( row=row, column=0, padx=5 )
            bar.grid( row=row, column=1, padx=5 )
            text    = Tk.Label( progress_table, textvariable=tim_var, justify=Tk.RIGHT )
            text.grid( row=row, column=2, padx=5 )
            text    = Tk.Label( progress_table, textvariable=max_var, justify=Tk.RIGHT )
            text.grid( row=row, column=3, padx=5 )
            text    = Tk.Label( progress_table, textvariable=val_var, justify=Tk.RIGHT )
            text.grid( row=row, column=4, padx=5 )
            blink.grid( row=row, column=5 )

        row += 1
        self.total_tim_lbl = Tk.StringVar()
        label   = Tk.Label( progress_table, textvariable=self.total_tim_lbl, justify=Tk.RIGHT )
        label.grid( row=row, column=1, sticky=Tk.E )
        self.total_tim_var = Tk.StringVar()
        text    = Tk.Label( progress_table, textvariable=self.total_tim_var, justify=Tk.RIGHT )
        text.grid( row=row, column=2, padx=5 )

        if self.plot_init_cb is not None:
            self.plot_init_cb( self, iframe )

        log_frame = Tk.Frame( iframe )
        log_frame.pack( fill=Tk.BOTH, expand=1 )
        label = Tk.Label( log_frame, text="Execution Log" )
        label.pack( anchor=Tk.NW )

        self.log_text = ScrolledText( log_frame, width=100, height=20 )
        self.log_text.pack( fill=Tk.BOTH )
        for tag, color in TAG_COLORS.items():
            self.log_text.tag_configure( tag, foreground=color )

    def buttonbox(self):
        '''
        override standard buttonbox.
        add "Default" button
        '''

        box = Tk.Frame(self)

        self.button = Tk.Button(box, text="Cancel", width=10, command=self.button_command)
        self.button.pack(side=Tk.LEFT, padx=5, pady=10)

        self.bind("<Escape>", self.button_command)
        box.pack()

    def cancel( self ):
        """
            We have to override Dialog.cancel because the following
            call is implemented in Dialog.__init__

            self.protocol( "WM_DELETE_WINDOW", self.cancel )
        """
        self.button_command()

    def button_command( self, *args ):
        btn_text = self.button.cget( 'text' )
        if btn_text == 'OK':
            self.terminate( btn_text )
            return

        yn = MessageBox.askyesno(
                'Cancel Confirmation',
                'Do you really want to stop the execution?',
                parent=self,
                )
        if yn:
            self.stop_mode = True
            parent_ = self.parent
            self.terminate( btn_text )
            MessageBox.showinfo(
                'Cancel Notification',
                'Process has been canceled.',
                parent=parent_,
                )

    def terminate( self, btn_text ):
        if self.terminate_cb is not None:
            self.terminate_cb( btn_text )
        self.show_return = self.button.cget( 'text' )
        self.destroy()

    def get_log_text(self, fmt='%(asctime)s %(message)s', datefmt='%H:%M:%S'):
        """ Convert log messages to text. On formats see logger.Formatter """

        if len( self.log_records ) == 0:
            return None

        f   = logging.Formatter(fmt, datefmt)
        buf = StringIO()
        b   = 0
        warn_stat = 0
        warn_end = 0
        max_level = None

        while True:
            try:
                record = self.log_records.pop(0)
            except IndexError:
                # stop if there exist no records left
                break

            if max_level is None or record.levelno > max_level:
                max_level = record.levelno
            if b:
                b += buf.write('\n')
                if record.levelno >= logging.WARNING:
                    if warn_stat == 0:
                        warn_stat = b
                    warn_end = b

            try:
                formated_record = f.format(record)
            except:
                from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
                etb = ExceptionTracebacker()
                formated_record = 'format failed with %s "%s"\n' % (etb.last_lines(), str(record))
            b += buf.write(formated_record)
            if max_level >= logging.WARNING:
                warn_end = b
                # stop once in order to avoid coloring of normal messages
                break

        text = buf.getvalue()
        buf.close()

        return text, max_level, warn_stat, warn_end

    def refresh_log( self ):
        ret = self.get_log_text()
        if ret is None:
            return

        text, level, warn_stat, warn_end = ret
        # print( 'warn_stat, warn_end=', warn_stat, warn_end )

        if self.refresh_log_cb is not None:
            self.refresh_log_cb( text, level )

        for key_word, key_level in [    [ 'error',      logging.ERROR ],
                                        [ 'warning',    logging.WARNING ],
                                        ]:
            if text.lower().find( key_word ) >= 0:
                level = max( level, key_level )

        if level >= logging.ERROR:
            self.pinfo.set_error()

        has_error_or_warinng = level >= logging.WARNING
        log_text = self.log_text
        log_text.configure( state=Tk.NORMAL )

        if has_error_or_warinng:
            # reduce the index by one, e.g., 13.0 ==> 12.0
            msg_start = re.sub( r'(\d+).', lambda m: '%d.' % ( int( m.group(1) ) - 1 ), log_text.index( Tk.END ) )

        log_text.insert( Tk.END, text )

        if has_error_or_warinng:
            msg_end = log_text.index( Tk.END )

        log_text.insert( Tk.END, '\n' )

        if has_error_or_warinng:
            tag = TEXT_TAGS[level]
            if warn_stat > 0:
                # modify msg_start, msg_end
                t1 = text[warn_stat:warn_end]
                t2 = text[warn_end:]
                n_ = [ len( [pos for pos, char in enumerate(t_) if char == '\n'] ) for t_ in [t1, t2 ] ]
                n1 = n_[0] + n_[1]
                n2 = n_[1]
                msg_start = re.sub( r'(\d+).', lambda m: '%d.' % ( int( m.group(1) ) - (n1+1) ), msg_end )
                if n2 > 0:
                    msg_end = re.sub( r'(\d+).', lambda m: '%d.' % ( int( m.group(1) ) - (n2+1) ), msg_end )
                # print( 'n_=', n_, 'msg_start, msg_end=', (msg_start, msg_end) )
            log_text.tag_add( tag, msg_start, msg_end )

        log_text.see( Tk.END )

        log_text.configure( state=Tk.DISABLED )

    def refresh( self, terminate_loop=False ):
        self.refresh_log()

        info = self.pinfo.get_info()

        current_stream = self.pinfo.current_stream
        todo = self.pinfo.stream_max_dict[current_stream]
        if todo == 0:
            # update time only for todo > 0
            self.after( self.refresh_interval, self.refresh )
            return

        start_time = self.pinfo.get_start_time(current_stream)
        this_time = time.time()
        if start_time is None:
            disp_time = ''
        else:
            disp_time = make_disp_time( this_time - start_time )
        self.vars_[current_stream][0].set( disp_time )
        if current_stream > self.first_stream:
            start_time = self.pinfo.get_start_time(self.first_stream)
            if start_time is not None:
                self.total_tim_lbl.set( 'cumulative time' )
                self.total_tim_var.set( make_disp_time( this_time - start_time ) )

        if info is None:
            num_stopped = 0
            for state in self.button_states:
                num_stopped += state
            if self.num_buttons > num_stopped:  # namely, if there remains an un-finished button
                self.after( self.refresh_interval, self.refresh )
            else:
                if not self.can_safely_finish:
                    self.after( self.refresh_interval, self.refresh )
                # better termination test desired including minor error cases
                pass
            return

        progresses = self.pinfo.get_stream_progresses()

        if False:
            logger = logging.getLogger( __name__ )
            logger.info( 'progresses=' + str(progresses) )

        for i, var_row in enumerate( self.vars_ ):
            _, max_var, val_var, stt_var = var_row
            progress = progresses[i]
            if progress[0]  == 0:
                # if todo == 0 then skip
                continue

            max_var.set( str( progress[0] ) )
            val_var.set( str( progress[1] ) )
            state = progress[2]
            stt_var.set( STATE_TEXTS[state] )

            label   = self.blinks[i][0]
            blink   = self.blinks[i][1]
            if state == STATE_DOING:
                if not blink.switch:
                    blink.start()

                if label.cget('fg') == 'red':
                    label.config( fg=GREEN )
                elif label.cget('bg') == 'red':
                    label.config( bg=GREEN )

                if self.is_alive_cb is not None:
                    is_alive = self.is_alive_cb()
                    is_alive = False
                    if self.worker_is_alive and not is_alive:
                        self.found_inactive_time = time.time()
                        self.worker_is_alive = False
                    if not is_alive:
                        t = time.time() - self.found_inactive_time
                        if t > 10:
                            state = STATE_DONE
                            if self.logger is None:
                                logger = logging.getLogger(__name__)
                            else:
                                logger = self.logger
                            logger.warning("state has been forced to be STATE_DONE after %.3g inactive time", t)

            if state in [ STATE_DONE, STATE_ERROR, STATE_FATAL ]:
                blink.stop()
                self.button_states[i] = 1       # this terminates the after loop when all the button states are 1

                if state == STATE_DONE:
                    label.config( bg=GREEN )
                    label.config( fg='white' )
                else:
                    label.config( bg='red' )
                    label.config( fg='white' )
                label.update()

            elif state == SAFE_FINISH:
                self.can_safely_finish = True

            bar = self.bars[i]
            bar['maximum']  = progress[0]
            bar['value']    = min( progress[0], progress[1] )
        states = self.pinfo.get_stream_states()

        if self.plot_update_cb is not None:
            self.plot_update_cb( self )

        if self.pinfo.is_end_states( states ):
            self.is_at_end_ = True
            self.button.config( text='OK' )
            self.bind("<Return>", self.button_command)
            if self.plot_final_cb is not None:
                self.plot_final_cb()
            terminate_next = True
            # there exist remaining log texts which have not yet been shown
        else:
            terminate_next = False

        if not terminate_loop:
            self.after( self.refresh_interval, lambda: self.refresh(terminate_loop=terminate_next) )

    def is_at_end( self ):
        return self.is_at_end_
    


"""
    V2PropOptimizer.PropOptimizerEDM.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize, basinhopping
from time import sleep, time
from molass_legacy._MOLASS.SerialSettings import set_setting
from molass_legacy.Models.RateTheory.EDM import guess_multiple, edm_impl
from molass_legacy.KekLib.OurTkinter import Tk
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.SerialAnalyzer.ElutionCurve import ElutionCurve
from .PropOptimizer import compute_range_rgs, PROP_MIN_VALUE, RangeRgComputer

VERY_SMALL_RG = 1

def guess_init_params(x, y):
    num_peaks = 2
    params_array = np.array(guess_multiple(x, y, num_peaks))
    return params_array

def compute_props(x, params_array):
    areas = []
    for params in params_array:
        cy = edm_impl(x, *params)
        areas.append(np.sum(cy))
    props = np.array(areas)/np.sum(areas)
    return props

def compute_cy_list(x, params_array):
    cy_list = []
    for params in params_array:
        cy = edm_impl(x, *params)
        cy_list.append(cy)
    return cy_list

def optimize_to_props(x, y, peaks, props, use_basinhopping=False):
    assert len(peaks) == len(props)
    peaks = np.asarray(peaks)
    props = np.asarray(props)

    def objective(p):
        ty = np.zeros(len(y))
        p_ = p.reshape(peaks.shape)
        areas = []
        try:
            cy_list = []
            for params in p_:
                cy = edm_impl(x, *params)
                cy_list.append(cy)
                areas.append(np.sum(cy))
                ty += cy
            props_ = np.array(areas)/np.sum(areas)
            area_dev = np.sum((ty - y)**2)
            prop_dev = max(PROP_MIN_VALUE, np.sum((props_ - props)**2))
            fv = np.log(area_dev) + np.log(prop_dev) - np.log(np.std(p_[:,1]))
        except:
            fv = np.inf
        return fv

    init_params = peaks.flatten()
    if use_basinhopping:
        ret = basinhopping(objective, init_params, minimizer_kwargs=dict(method="Nelder-Mead"))
    else:
        ret = minimize(objective, init_params, method="Nelder-Mead")
    return ret

def compute_optimal_proportion(progress_queue, job_args):
    print("compute_optimal_proportion EDM")

    progress = 0

    root = Tk.Tk()

    v2_optimizer = job_args.v2_optimizer
    xr_curve = v2_optimizer.xr_curve
    x = xr_curve.x
    y = xr_curve.y
    num_peaks = 2

    params_array = np.array(guess_multiple(x, y, num_peaks))

    def plot_model(params_array, title, rgs=None):
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_title(title)
            ax.plot(x, y, color="orange")
            cy_list = []
            areas = []
            for params in params_array:
                cy = edm_impl(x, *params)
                cy_list.append(cy)
                areas.append(np.sum(cy))
                ax.plot(x, cy, ":")
            ty = np.sum(cy_list, axis=0)
            ax.plot(x, ty, ":", color="red")
            props = areas/np.sum(areas)
            xmin, xmax = ax.get_xlim()
            ymin, ymax = ax.get_ylim()
            tx = xmin*0.8 + xmax*0.2
            ty = ymin*0.3 + ymax*0.7
            ax.text(tx, ty, "props=%.3g, %.3g" % tuple(props))

            if rgs is not None:
                ty = ymin*0.5 + ymax*0.5
                ax.text(tx, ty, "rgs=%.3g, %.3g" % tuple(rgs))

            fig.tight_layout()
            plt.show()

    progress += 1
    progress_queue.put([progress])
    plot_model(params_array, "after guess_multiple")

    p = job_args.init_prop
    init_pv = np.array((p, 1-p))

    def objective_prop(params_vector, target_pv, return_full=False):
        cy_list = []
        areas = []
        for params in params_vector.reshape(params_array.shape):
            cy = edm_impl(x, *params)
            cy_list.append(cy)
            areas.append(np.sum(cy))
        ty = np.sum(cy_list, axis=0)
        props = np.array(areas)/np.sum(areas)
        fv = np.log(np.sum((ty - y)**2))*0.4 + np.log(max(PROP_MIN_VALUE, np.sum((props - target_pv)**2)))*0.6
        if return_full:
            return fv, cy_list
        else:
            return fv

    prop_ret = minimize(lambda x: objective_prop(x, init_pv), params_array.flatten(), method="Nelder-Mead")

    progress += 1
    progress_queue.put([progress])
    plot_model(prop_ret.x.reshape(params_array.shape), "after minimize objective_prop")

    ecurve = ElutionCurve(y)
    paired_ranges = ecurve.get_default_paired_ranges()
    print("paired_ranges=", paired_ranges)
    qv = v2_optimizer.qvector
    D = v2_optimizer.xrD
    E = v2_optimizer.xrE

    init_params = np.concatenate([[job_args.init_prop], prop_ret.x])

    # maxiter = len(init_params)*200
    maxiter = len(init_params)*20
    call_count = 0
    temp_progress = None

    rg_computer = RangeRgComputer(qv, D, E, paired_ranges)

    def objective_rdr(params, return_details=False):
        nonlocal call_count, temp_progress

        call_count += 1
        if call_count % 100 == 0:
            print("called", call_count)
            ratio = call_count/maxiter
            temp_progress = progress + int((10 - progress)*ratio)
            progress_queue.put([temp_progress])

        p = params[0]
        target_pv = np.array((p, 1-p))
        prop_fv, cy_list = objective_prop(params[1:], target_pv, return_full=True)
        C = np.array(cy_list)
        rgs = rg_computer.compute(C)

        denom = rgs[0] + rgs[1]
        if denom < VERY_SMALL_RG:
            rdr = 1
        else:
            rdr = abs(rgs[0] - rgs[1])/denom

        if return_details:
            return np.log(rdr), prop_fv, rgs
        else:
            return np.log(rdr)*0.25 + prop_fv*0.75

    t0 = time()
    rdr_ret = minimize(objective_rdr, init_params, method="Nelder-Mead", options=dict(maxiter=maxiter))
    print("it took", time() - t0)

    progress = temp_progress
    progress += 1
    progress_queue.put([progress])

    fv1, fv2, rgs = objective_rdr(rdr_ret.x, return_details=True)
    print("fv1=", fv1, "fv2=", fv2, "rgs=", rgs)
    plot_model(rdr_ret.x[1:].reshape(params_array.shape), "after minimize objective_rdr", rgs=rgs)

    progress_queue.put([-1])



# coding: utf-8
"""
    ConnectedIntervals.py

    連結区間の識別

    Copyright (c) 2016-2017, Masatsuyo Takahashi, KEK-PF
"""

ALLOWANCE   = 10

class ConnectedIntervals:
    def __init__( self, index_array ):
        intervals = []
        last = None
        connected_interval = []
        for i in index_array:
            if not (last is None or ( i - last ) < ALLOWANCE ):
                intervals.append( connected_interval )
                connected_interval = []

            connected_interval.append( i )
            last = i

        intervals.append( connected_interval )
        self.intervals  = intervals



"""
    Optimizer.Devel.SimpleInspect.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import os
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Optimizer.BasicOptimizer import UV_XR_RATIO_ALLOW

def simple_inspect_impl(optimizer, work_params):
    from importlib import reload
    import Optimizer.XrUvScaleRatio
    reload(Optimizer.XrUvScaleRatio)
    from molass_legacy.Optimizer.XrUvScaleRatio import xruv_scale_ratio_penalty
    print("simple_inspect_impl")
    xr_params, xr_baseparams, rg_params, (a, b), uv_params, uv_baseparams, (c, d), sdmcol_params = optimizer.split_params_simple(work_params)

    print("real_bounds=", optimizer.real_bounds)

    uv_xr_ratio = uv_params/xr_params
    average = np.average(uv_xr_ratio)
    std = np.std(uv_xr_ratio)
    ratio_deviation = max(0, std/average - UV_XR_RATIO_ALLOW)**2
    print("np.std(uv_xr_ratio) = ", std)
    print("np.average(uv_xr_ratio) = ", average)
    print("ratio_deviation = ", ratio_deviation)
    print("penalty = ", xruv_scale_ratio_penalty(xr_params, uv_params))

    with plt.Dp():
        fig, ax = plt.subplots()
        ax.plot(uv_xr_ratio, "o")
        ax.axhline(np.average(uv_xr_ratio), color="red")
        ax.axhspan(average - std, average + std, color="red", alpha=0.2)
        fig.tight_layout()
        plt.show()


# coding: utf-8
"""
    SimpleGuinierAnalyzer.py

    Copyright (c) 2017-2018, Masatsuyo Takahashi, KEK-PF
"""
import re
import numpy                as np
import matplotlib.pyplot    as plt
import matplotlib.gridspec  as gridspec
from mpl_toolkits.mplot3d   import Axes3D
import seaborn
from CanvasDialog           import CanvasDialog

NUM_ZX_POINTS   = 5

doing_re            = re.compile( r'Doing (\S+)' )
guinier_start_re    = re.compile( r'^guinier start' )
guinier_result_re   = re.compile( r'^guinier result: type=(\S+), Rg=\((\S+),\s*\S+\), basic_quality=(\S+)' )
extrapolation_re    = re.compile( r'^extrapolation start for the (\d)-th peak ad\((\d)\) with range[\(,\[](\d+), (\d+), (\d+)[\),\]]' )
result_type_re      = re.compile( r'(.+)\((.+)\)' )

class SimpleGuinierAnalyzer:
    def __init__( self, tester_log ):
        self.result_dict = {}
        self.result_type_dict = {}
        self.renumd_type_dict = {}
        self.folder_no = -1
        self.read_log( tester_log )
        self.create_xyz()

    def result_type_to_chart_value( self, type_ ):
        m = result_type_re.match( type_ )
        if m:
            branch  = m.group(1)
            try_ret = m.group(2)

            if branch == 'None':
                v = -20
            else:
                digits = [ int(d) for d in branch.split( '-' ) ]
                lacking = 5 - len(digits)
                if lacking > 0:
                    digits += [0] * lacking

                v = 0
                for d in digits:
                    v = v*2 + d

                v *= 20
                if try_ret != 'None':
                    v += int(try_ret)

        else:
            assert False

        self.result_type_dict[v] = type_

        return v

    def read_log( self, tester_log ):

        self.folder_no = -1
        self.folder_paths = []
        self.folder_infos = []

        fh = open( tester_log )

        # for cases where guinier_start_re does not match first
        before_guinier_start = True

        self.max_row_no = None
        folder_peaks = None

        for line in fh:
            m = doing_re.match( line )
            if m:
                if folder_peaks is not None:
                    self.folder_infos.append( folder_peaks )

                folder = m.group(1)
                self.folder_no += 1
                print( self.folder_no, folder )
                self.folder_paths.append( folder )
                row_no  = -1
                folder_peaks = []
                continue

            m = guinier_start_re.match( line )
            if m:
                zx_type = None
                zx_count = None
                file_no = -1
                before_guinier_start = False

            if before_guinier_start:
                # ignore results before guinier_start.
                # this is normally considered unnecessary,
                # but it did happen once for 20180204/Open01, though not reproducibly
                continue

            m = guinier_result_re.match( line )
            if m:
                type_   = m.group(1)
                rg      = m.group(2)
                bq      = m.group(3)
                row_no += 1
                if self.max_row_no is None or row_no > self.max_row_no:
                    self.max_row_no = row_no

                if zx_type is None:
                    file_no += 1
                else:
                    zx_count += 1
                    if zx_type == 0:
                        if zx_count < NUM_ZX_POINTS:
                            file_no = None
                        else:
                            file_no = zx_start + zx_count
                    else:
                        num_remain = ( zx_end - zx_start ) - zx_count
                        if num_remain >= 0:
                            file_no = zx_start - zx_count
                        else:
                            file_no = None
                print( row_no, file_no, zx_count, type_, rg, bq )
                file_id = str(file_no) + '-' + str(zx_count)
                chart_value = self.result_type_to_chart_value( type_ )
                self.result_dict[ (row_no, self.folder_no) ] = [ file_id, chart_value, rg, bq ]
                continue

            m = extrapolation_re.match( line )
            if m:
                type_   = None
                peak    = int( m.group(1) )
                ad      = int( m.group(2) )
                left    = int( m.group(3) )
                middle  = int( m.group(4) )
                right   = int( m.group(5) )
                row_no += 1
                if ad == 0:
                    folder_peaks.append( ( left, middle, right ) )
                print( row_no, None, type_, peak, ad )
                zx_type = ad
                zx_count    = -1
                if ad == 0:
                    zx_start    = left
                    zx_end      = middle
                else:
                    zx_start    = middle
                    zx_end      = right

        if folder_peaks is not None:
            self.folder_infos.append( folder_peaks )

        fh.close()

        print( 'max_row_no=', self.max_row_no )
        print( 'folder_no=', self.folder_no )

        self.type_names = []
        for n, v in enumerate( sorted( self.result_type_dict.keys() ) ):
            type_name = self.result_type_dict[v]
            self.renumd_type_dict[v] = [ n, type_name ]
            self.type_names.append( type_name )

    def write_csv( self ):
        fh_all  = open( 'result_summary_all.csv', 'w' )
        fh_id   = open( 'result_summary_id.csv', 'w' )
        fh_type = open( 'result_summary_type.csv', 'w' )
        fh_rg   = open( 'result_summary_rg.csv', 'w' )
        fh_bq   = open( 'result_summary_bq.csv', 'w' )

        hf_list = [ fh_id, fh_type, fh_rg, fh_bq ]

        header = ','.join( [ 'row no' ] + [ 'folder-%03d' % n for n in range( self.folder_no+1 ) ] ) + '\n'
        fh_all.write( header )
        for fh in hf_list:
            fh.write( header )

        for row in range( self.max_row_no + 1 ):
            result_row = []
            for col in range( self.folder_no + 1 ):
                result = result_dict.get( (row, col) )
                if result is None:
                    result_row.append( str(result) )
                else:
                    result[1] = renumd_type_dict[ result[1] ][0]
                    result_row.append( ':'.join( [ str(s) for s in result ] ) )

            fh_all.write( ','.join( [ str(row) ] + result_row ) + '\n' )

            for n in range(len(hf_list)):
                fh = hf_list[n]
                fh.write( ','.join( [ str(row) ] + [ '' if result == 'None' else str(result.split(':')[n]) for result in result_row ] ) + '\n' )

        fh_all.close()
        for fh in hf_list:
            fh.close()

    def create_xyz( self ):
        x_list = []
        y_list = []
        z_list = []
        r_list = []

        for col in range( self.folder_no + 1 ):
            for row in range( self.max_row_no + 1 ):
                result = self.result_dict.get( (row, col) )
                if result is None:
                    continue

                type_val    = self.renumd_type_dict[ result[1] ][0]
                x_list.append( row )
                y_list.append( col )
                z_list.append( type_val )
                rg = result[2]
                r_list.append( 0 if rg =='None' else float(rg)  )

        self.x = np.array( x_list )
        self.y = np.array( y_list )
        self.z = np.array( z_list )
        self.r = np.array( r_list )

    def plot_results( self, fig ):
        print( 'plot_results' )
        # fig = plt.figure( figsize=(18, 9) )
        gs  = gridspec.GridSpec( 1, 1 )
        self.ax1 = ax1 = fig.add_subplot( gs[0,0], projection='3d' )
        self.plot_redraw( fig, set_lims=False )
        self.xlim = np.array( ax1.get_xlim() )  # deep copy
        self.ylim = np.array( ax1.get_ylim() )  # deep copy
        self.zlim = np.array( ( 0, 200 ) )
        ax1.set_xlim( self.xlim )
        ax1.set_ylim( self.ylim )
        ax1.set_zlim( self.zlim )

        fig.tight_layout()

    def plot_redraw( self, fig, draw_flags=None, set_lims=True ):
        ax1 = self.ax1
        ax1.cla()
        ax1.set_title( 'Rg and type of peak processing' )
        ax1.set_xlabel( 'File No' )
        ax1.set_ylabel( 'Folder No' )
        ax1.set_zlabel( 'Rg' )

        max_z = np.max( self.z )

        for n in range(max_z + 1):
            if draw_flags is not None:
                if draw_flags[n] == 0:
                    continue

            index = self.z == n
            ax1.scatter( self.x[index], self.y[index], self.r[index], s=3, label='type %d' % n )

        if set_lims:
            ax1.set_xlim( self.xlim )
            ax1.set_ylim( self.ylim )
            ax1.set_zlim( self.zlim )
        ax1.legend()

    def show_dialog( self, parent ):
        dialog = CanvasDialog( "Debug", parent=parent )
        dialog.show( self.plot_results, figsize=(18,9), toolbar=True )



# coding: utf-8
"""
    SimpleGuinierTester.py

    Copyright (c) 2017, Masatsuyo Takahashi, KEK-PF
"""
import re
from time           import sleep
from molass_legacy.KekLib.OurTkinter     import Tk
from TkTester       import TestClient, reply_ok, reply_yes, reply_no, messagebox_shown
from DataUtils      import serial_folder_walk

class SimpleGuinierTester:
    def __init__( self, app ):
        self.app        = app

    def control_suite( self, root_folder, targets={} ):
        self.root_folder    = root_folder
        self.compiled_targets   = {}
        for k, v in targets.items():
            regex = re.compile( k )
            self.compiled_targets[ regex ] = v

        self.results    = []
        self.test_count = 0
        self.ok_count   = 0
        self.client = TestClient( self.app, self.tester_cb )

        sleep( 1 )

        print( 'control_suite done' )

    def tester_cb( self, client, agent ):
        self.client = client
        self.agent  = agent
        print( 'tester_cb: root_folder=', self.root_folder )

        serial_folder_walk( self.root_folder, self.do_a_folder )

        # agent.quit( immediately=True )    # does not work
        # self.app.quit( immediately=True ) # does not work

        self.app.parent.quit()              # does work
        print( 'test results' )
        for result in self.results:
            print( result )

        print( 'test_count=', self.test_count )
        if self.ok_count == self.test_count:
            print( 'All tests were successful.' )
        else:
            failures = self.test_count - self.ok_count
            print( 'There were %d failures.' %  failures )

    def get_expected_results( self, in_folder ):

        ret = None
        for k, v in self.compiled_targets.items():
            if k.search( in_folder ):       # changed from find to re.search
                ret = v
                break

        return ret

    def do_a_folder( self, in_folder, uv_folder, plot=None, suppress_retry_call=False, analysis_name=None ):

        expected = self.get_expected_results( in_folder )
        if expected is None:
            return True, None

        print( 'expected=', expected )

        agent   = self.agent

        print( 'in_folder=', in_folder )
        print( 'uv_folder=', uv_folder )
        agent.in_folder_entry.focus_force()
        print( 'in_folder_entry.focus_force' )
        agent.in_folder_entry.delete( 0, Tk.END)
        agent.in_folder_entry.insert( 0, in_folder )
        agent.on_entry_in_folder()

        if uv_folder != in_folder:
            agent.uv_folder_entry.focus_force()
            agent.uv_folder_entry.delete( 0, Tk.END)
            agent.uv_folder_entry.insert( 0, uv_folder )
            agent.on_entry_uv_folder()

        for k, v in sorted( expected.items() ):
            self.test_count += 1
            try:
                result = agent.do_simple_guinier( k )
                ok = abs( result.Rg - v )/v < 0.01
                if ok:
                    self.ok_count += 1
                rg = result.Rg
            except:
                ok = None
                rg = None

            self.results.append( [ in_folder, k, v, float('%.3g' % rg), ok ] )

        sleep( 1 )

        ret_continue = True
        return ret_continue, None


"""
    Trimming.OptViewRange.py

    Copyright (c) 2024-2025, SAXS Team, KEK-PF
"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.Batch.LiteBatch import LiteBatch
from molass_legacy.Peaks.MomentsUtils import compute_moments
from molass_legacy.DataStructure.LPM import get_corrected

BAD_SN_NEGATIVE_THRESHOLD = 0.3

def correct_largely_negative_values(y, num_peaks):
    y = get_corrected(y)
    where_negetive = np.where(y < 0)[0]
    negetve_ratio = len(where_negetive)/len(y)
    print("negetve_ratio=", negetve_ratio)
    if num_peaks == 1 or negetve_ratio > BAD_SN_NEGATIVE_THRESHOLD:
        # as in 20250212/REF
        import logging
        logger = logging.getLogger(__name__)
        logger.warning("applied correct_largely_negative_values due to negetve_ratio=%.3g", negetve_ratio)          
        m = np.argmax(y)
        y_ = y.copy()
        left_negative = where_negetive[where_negetive < m]
        if len(left_negative) > 0:
            y_[:left_negative[-1]] = 0
        right_negative = where_negetive[where_negetive > m]
        if len(right_negative) > 0:
            y_[right_negative[0]:] = 0
    else:
        y_ = y
    return y_

class OptViewRange:
    def __init__(self, x, y, num_peaks, upper=None, smooth=False):
        if smooth:
            from molass_legacy.KekLib.SciPyCookbook import smooth
            y = smooth(y)
        if upper is None:
            x_ = x
            y_ = y
        else:
            y = correct_largely_negative_values(y, num_peaks)
            max_y = np.max(y)
            upper_part = y > max_y * upper
            x_ = x[upper_part]
            y_ = y[upper_part]
        self.x_ = x_
        self.y_ = y_
        self.moments = compute_moments(x_, y_)

    def get_range(self, sigmas=5):
        m1 = self.moments[1]
        hw = sigmas * np.sqrt(self.moments[2])
        return m1 - hw, m1 + hw
    
def get_opt_view_range(caller, debug=True):
    print("get_opt_view_range")
    for keystr in ["uv_restrict_list", "xr_restrict_list"]:
        print(keystr, get_setting(keystr))

    lb = LiteBatch()
    lb.prepare(caller.serial_data, debug=False)
    uv_x, uv_y, xr_x, xr_y, baselines = lb.get_curve_xy(return_baselines=True, debug=False)
    uv_y_ = uv_y - baselines[0]
    xr_y_ = xr_y - baselines[1]
    ranges = []
    ranges_xy = []
    for x, y in [(uv_x, uv_y_), (xr_x, xr_y_)]:
        range_ = OptViewRange(x, y, upper=0.03)
        ranges.append(range_.get_range())
        ranges_xy.append((range_.x_, range_.y_))

    if debug:
        with plt.Dp():
            fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
            ax1.plot(uv_x, uv_y)
            ax1.plot(uv_x, baselines[0])
            ax1.plot(*ranges_xy[0], 'o', color='yellow')
            ax1.axvspan(*ranges[0], alpha=0.2)
            ax2.plot(xr_x, xr_y)
            ax2.plot(xr_x, baselines[1])
            ax2.plot(*ranges_xy[1], 'o', color='yellow')
            ax2.axvspan(*ranges[1], alpha=0.2)
            fig.tight_layout()
            plt.show()


"""
    Theory.JsPedersen1997.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.integrate import quad, dblquad
from scipy.optimize import minimize
from numba.core.decorators import jit
import molass_legacy.KekLib.DebugPlot as plt

def F1(q, R):
    qR = q*R
    return 3*(np.sin(qR) - qR*np.cos(qR))/qR**3

def r5(R, e, x):
    return R*np.sqrt(np.sin(x)**2 + e**2*np.cos(x)**2)

# @jit(nopython=True)
def P5(qv, R, e):
    I = np.zeros(len(qv))
    for i, q in enumerate(qv):
        def f(x):
            return F1(q, r5(R, e, x))**2 * np.sin(x)
        I[i] = quad(f, 0, np.pi/2)[0]
    return I

def r6(a, b, c, alpha, beta):
    return np.sqrt((a**2*np.sin(beta)**2 + b**2*np.cos(beta)**2)*np.sin(alpha)**2 + c**2*np.cos(alpha)**2)

# @jit(nopython=True)
def P6(qv, a, b, c):
    I = np.zeros(len(qv))
    for i, q in enumerate(qv):
        def f(x, y):
            return F1(q, r6(a, b, c, x, y))**2 * np.sin(x)
        I[i] = dblquad(f, 0, np.pi/2, lambda x: 0, lambda x: np.pi/2)[0]
    return 2/np.pi*I

def demo():
    qv = np.linspace(0.005, 0.5, 100)

    fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(16,7))

    ax1.set_title("Ellipsoid of Revolution", fontsize=16)
    ax2.set_title("Tri-axial Ellipsoid", fontsize=16)

    R = 40
    for ax in [ax1, ax2]:
        ax.set_yscale('log')
        ax.plot(qv, F1(qv, R)**2, label='Sphere')

    for e in [0.9, 1, 1.1]:
        ax1.plot(qv, P5(qv, R, e), ':', label='Ellipsoid e=%g' % e)

    for a, b, c in [(R, R, R)]:
        ax2.plot(qv, P6(qv, a, b, c), ':', label='Ellipsoid a,b,c=%s' % str((a,b,c)))

    for ax in [ax1, ax2]:
        ax.legend()

    fig.tight_layout()
    plt.show()

class SolidEllipsoidRev:
    def __init__(self):
        pass

    def fit(self, qv, Ic, initRg):
        scale = np.sqrt(5/3)
        initR = scale*initRg

        def obj_func(p):
            C, R, e = p
            return np.sum((Ic - C*P5(qv, R, e))**2)

        res = minimize(obj_func, (1, initR, 1))
        self.C, self.R, self.e = res.x
        self.Rg = self.R*np.sqrt((1+self.e**2)/4)
        return res.x

    def intensity(self, qv):
        return self.C * P5(qv, self.R, self.e)

class TriAxialEllipsoid:
    def __init__(self):
        pass

    def fit(self, qv, Ic, initRg):
        scale = np.sqrt(5/3)
        initR = scale*initRg

        def obj_func(p):
            C, a, b, c = p
            return np.sum((Ic - C*P6(qv, a, b, c))**2)

        res = minimize(obj_func, (1, initR, initR, initR))
        self.C, self.a, self.b, self.c = res.x
        self.Rg = np.sqrt((self.a**2 + self.b**2 + self.c**2)/5)
        return res.x

    def intensity(self, qv):
        return self.C * P6(qv, self.a, self.b, self.c)


def demo2():

    Rg = 35
    qv = np.linspace(0.005, 0.4, 100)

    fig, ax = plt.subplots()

    ax.set_title("Various Shapes of Ellipsoids with the same Rg=35", fontsize=20)
    ax.set_yscale('log')

    R = np.sqrt(5/3)*Rg
    b1 = np.pi/R
    b2 = b1*1.5

    ax.plot(qv, F1(qv, R)**2, label='Sphere')

    for a, b, c in np.random.uniform(0, 1, (10, 3)):
        print((a, b, c))
        r = np.sqrt((a**2 + b**2 + c**2)/5)
        scale = Rg/r
        a_ = a*scale
        b_ = b*scale
        c_ = c*scale
        ax.plot(qv, P6(qv, a_, b_, c_), label='Ellipsoid(%.3g, %.3g, %.3g)' % (a_, b_, c_))

    ymin, ymax = ax.get_ylim()
    ax.set_ylim(ymin, ymax)
    for b, pc in zip([b1, b2], [50, 1]):
        ax.plot([b, b], [ymin, ymax], ':', color='red', label='rank2 ratio %d%%' % pc)

    ax.legend()
    fig.tight_layout()
    plt.show()



"""
    Trimming.FlangeLimit.py

    Copyright (c) 2018-2025, SAXS Team, KEK-PF
"""
import numpy as np
import logging
from bisect import bisect_right
from molass_legacy.KekLib.GeometryUtils import rotated_argmin

MIN_Q_LIMIT             = 0.2       #
MIN_Q_LIMIT_RATIO       = 0.67      # to cover 20161205
ALLOWANCE_WIDTH_RATIO   = 0.01
Q_LIMIT_RATIOS          = [0.99, 0.97, 0.95, 0.93]
OK_RATIO_LIMIT          = 0.15      # > 0.12 for 20190309_1
RETRY_RATIO_LIMIT       = 0.3
MIN_PEAK_NEG_ALLOW      = -0.03     # < -0.01 for SUB_TRN1
MIN_PEAK_POS_LIMIT      = 0.01
MIN_PEAK_WIDTH_Q        = 0.003     # < 0.006 for AhRR
LIMIT_TYPE_FLANGE = 0
LIMIT_TYPE_WA_CORNER = 1
END_FIND_ROTATION       = -np.pi/4

class FlangeLimit:
    def __init__(self, data, error, e_curve, q_vector):
        self.logger = logging.getLogger( __name__ )
        self.data = data
        self.error = error
        self.e_curve = e_curve
        self.q_vector = q_vector
        self.limit = None
        self.limit_type = None

    def get_limit(self, debug=False):
        if self.limit is None:
            try:
                self.get_limit_impl(self.e_curve, self.data, self.error, self.logger, debug=debug)
            except:
                from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
                etb = ExceptionTracebacker()
                self.logger.warning("get_limit_impl failed." + etb.last_lines(5))
        return self.limit

    def get_limit_impl(self, e_curve, data, error, logger, debug=False):
        if debug:
            import molass_legacy.KekLib.DebugPlot as plt
            
        # self.q_vector[-1] == 0.194 in 20171203
        i = min(int(len(self.q_vector)*MIN_Q_LIMIT_RATIO), bisect_right(self.q_vector, MIN_Q_LIMIT))
        j = e_curve.get_primarypeak_i()
        w = int(data.shape[0]*ALLOWANCE_WIDTH_RATIO)
        hw = w//2
        jslice = slice(j-hw, j+hw)
        avg_yd = np.average(data[:,jslice], axis=1)
        avg_yd[avg_yd <= 0] = np.nan        # to avoid np.log10() invalid value warning 
        log_yd = np.log10(avg_yd)
        error_ = error[:,jslice].copy()

        zeros_fix = False
        if error_[-1,0] == 0:
            # error data bug fix
            error_[-1,:] = error_[-2,:]
            zeros_fix = True
        elif error_[0,0] == 0:
            error_[0,:] = error_[1,:]
            zeros_fix = True

        if zeros_fix:
            self.logger.warning("zeros at the angular ends of error data have been replaced with the adjacent values.")

        log_ye = np.log10(np.average(error_, axis=1))
        log_yeg = np.gradient(log_ye)
        gy = log_yeg[i:]
        agy = np.abs(gy)

        hw_ = None
        def find_limit_cadidate(limit_ratio):
            nonlocal hw_
            k = int(len(agy)*limit_ratio)
            pp = np.argpartition(agy, k)[k:]
            f = np.min(pp)
            if gy[f] > 0:
                hw_ = hw
            else:
                # search in wider range as in 20170301
                hw_ = min(f//2, hw*10)      # so that f-hw_ > 0 for 20190317
            neighbor = gy[f-hw_:f+hw_]
            assert len(neighbor) > 0
            m = np.argmax(neighbor)
            f = f - hw_ + m  # change f to the neighbor max
            safe_f = f - w
            return pp, f, safe_f

        def has_a_peak_of_some_width(sf):
            y = log_ye[sf:] - log_ye[sf]
            wneg = np.where(y < MIN_PEAK_NEG_ALLOW)[0]
            if debug:
                print('y=', y[:10])
                plt.push()
                fig, ax = plt.subplots()
                ax.plot(y)
                ax.plot(wneg, y[wneg], 'o', color='yellow')
                plt.show()
                plt.pop()
            if len(wneg) > 0:
                y_ = y[0:wneg[0]]
                wpos = np.where(y_ > MIN_PEAK_POS_LIMIT)[0]
                if len(wpos) > 0:
                    qw = self.q_vector[sf+wpos[-1]] - self.q_vector[sf+wpos[0]]
                    # print('y=', y[0:10], 'wpos=', wpos[0:10], 'qw=', qw)
                    ret_judge = qw > MIN_PEAK_WIDTH_Q
                else:
                    qw = 0
                    ret_judge = False
            else:
                ret_judge = True
            if not ret_judge:
                self.logger.info("flange limit candidate has been discarded due to peak width of %.3g at Q=%.3g", qw, self.q_vector[sf])
            return ret_judge

        for k, q_limit_ratio in enumerate(Q_LIMIT_RATIOS):
            try:
                pp, f, safe_f = find_limit_cadidate(q_limit_ratio)
            except AssertionError:
                # as in 20200121_1
                continue

            check_ratio = np.average(agy[safe_f-w:safe_f])/agy[f]
            is_limit = np.isfinite(agy[f]) and check_ratio < OK_RATIO_LIMIT and has_a_peak_of_some_width(i+safe_f)
            if debug:
                # agy[f] can be inf as in 20170226/Sugiyama
                print([k], (f, self.q_vector[f]), is_limit, check_ratio, agy[safe_f-hw:safe_f+hw], agy[f-2:f+3])
            if is_limit or check_ratio > RETRY_RATIO_LIMIT:
                break
            else:
                # as in HIF
                continue

        fmin = np.argmin(log_ye[i:])
        if not is_limit or fmin < safe_f:
            # fmin < safe_f as in 0171203
            wy = log_ye[i+fmin:]
            if len(wy) > 1:
                m = rotated_argmin(END_FIND_ROTATION, wy, debug)
                safe_f = fmin + m
            else:
                safe_f = fmin
            is_limit = True
            self.limit_type = LIMIT_TYPE_WA_CORNER
            self.logger.info("wide angle limit has been determined as the corner of rightmost error-upturn.")
        else:
            self.limit_type = LIMIT_TYPE_FLANGE

        f_ = i + f
        safe_f_ = i + safe_f

        if debug:
            print('pp=', pp)
            print('check_ratio=', check_ratio)
            from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder
            limit_type = "Flange Limit" if self.limit_type == LIMIT_TYPE_FLANGE else "Wide Angle Limit"

            in_folder = get_in_folder()
            plt.push()
            fig = plt.figure(figsize=(21, 7))
            ax1 = fig.add_subplot(131)
            ax2 = fig.add_subplot(132)
            ax3 = fig.add_subplot(133)
            axes = [ax1, ax2, ax3]

            fig.suptitle(limit_type + " Debug for " + in_folder, fontsize=20)
            ax1.set_title("Log10( intensity )", fontsize=16)
            ax2.set_title("Log10( error )", fontsize=16)
            ax3.set_title("Gradient of Log10( error )", fontsize=16)

            ax1.plot(self.q_vector, log_yd)
            ax2.plot(self.q_vector, log_ye)
            ax3.plot(self.q_vector, log_yeg)
            ax3.plot(self.q_vector[i+pp], log_yeg[i+pp], 'o', color='yellow')
            fx, fy = self.q_vector[f_], log_yeg[f_]
            ax3.plot(fx, fy, 'o', color='pink')
            xmin, xmax = ax3.get_xlim()
            w = xmax - xmin
            ymin, ymax = ax3.get_ylim()
            h = ymax - ymin

            h_ = h*0.2
            for x_ in [f-hw_, f+hw_]:
                ix_ = min(i+x_, len(self.q_vector)-1)
                ax3.plot(self.q_vector[[ix_, ix_]], [-h_, +h_], ':', color='pink')

            if is_limit:
                sfx, sfy = self.q_vector[safe_f_], log_yeg[safe_f_]
                ax3.plot(sfx, sfy, 'o', color='red')
                ax3.annotate(limit_type, xy=(sfx, sfy), xytext=(sfx-0.1*w, sfy+0.2*h), ha="right", arrowprops=dict(arrowstyle="-", color='k'))
                for ax in axes[0:2]:
                    ymin, ymax = ax.get_ylim()
                    ax.set_ylim()
                    ax.plot([sfx, sfx], [ymin, ymax], ':', color='red', label=limit_type)
                    ax.legend()

            else:
                ax3.annotate("not a Flange Limit", xy=(fx, fy), xytext=(fx, fy+0.2*h), ha="center", arrowprops=dict(arrowstyle="-", color='k'))

            fig.tight_layout()
            fig.subplots_adjust(top=0.88)
            plt.show()
            plt.pop()

        if is_limit:
            self.limit = safe_f_



# coding: utf-8
"""
    EFA.Demo.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np
# import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.patches import Polygon
from CorrectedData import CorrectedXray
import molass_legacy.KekLib.DebugPlot as plt
from MatrixData import simple_plot_3d

def demo(in_folder):
    print(in_folder)
    xr = CorrectedXray(in_folder)
    D = xr.data
    print(D.shape)
    print('D[0:3,0:5]=', D[0:3,0:5])
    E = np.average(xr.error, axis=1)
    print(E.shape)
    print('E[0:3]=', E[0:3])
    A = (D.T/E.T).T
    print(A.shape)
    print('A[0:3,0:5]=', A[0:3,0:5])
    U, s, VT = np.linalg.svd(A)
    print('s.shape=', s.shape)

    num_elutions = A.shape[1]
    num_sings = 3
    fwd_sing_array = np.zeros((num_elutions, num_sings))*np.nan
    bwd_sing_array = np.zeros((num_elutions, num_sings))*np.nan
    for k in range(num_elutions):
        _, fs, _ = np.linalg.svd(A[:,0:k+1])
        _, bs, _ = np.linalg.svd(A[:,-(k+1):])
        fs_ = fs[0:num_sings]
        bs_ = bs[0:num_sings]
        fwd_sing_array[k,0:len(fs_)] = fs_
        bwd_sing_array[-(k+1),0:len(bs_)] = bs_

    gs = GridSpec(3,3)
    fig = plt.figure(figsize=(21, 12))
    ax00 = fig.add_subplot(gs[0,0], projection='3d')
    ax01 = fig.add_subplot(gs[0,1], projection='3d')
    ax02 = fig.add_subplot(gs[0,2])
    ax10 = fig.add_subplot(gs[1,0])
    ax11 = fig.add_subplot(gs[1,1])
    ax12 = fig.add_subplot(gs[1,2])
    ax12t = ax12.twinx()
    ax20 = fig.add_subplot(gs[2,0])
    ax21 = fig.add_subplot(gs[2,1])
    ax22 = fig.add_subplot(gs[2,2])
    ax22t = ax22.twinx()
    xr.plot_3d(ax00)
    simple_plot_3d(ax01, A)
    ax02.plot(s[0:5], '-o')
    for i in range(num_sings):
        fy = fwd_sing_array[:,i]
        by = bwd_sing_array[:,i]
        for ax in [ax10, ax12, ax20, ax22]:
            ax.plot(fy, label='fwd s%d' % i)

        for ax in [ax11, ax12t, ax21, ax22t]:
            ax.plot(by, label='bwd s%d' % i)

    ylim_array = np.array([ax.get_ylim() for ax in [ax10, ax11, ax12]])
    ymin = min(ylim_array[:,0])
    ymax = max(ylim_array[:,1])

    ey = xr.ecurve.y / xr.ecurve.max_y * ymax
    ex = np.arange(len(ey))
    points = [(0,0)] + list(zip(ex,ey))+ [(ex[-1],0)]

    for ax in [ax10, ax11,ax12]:
        poly = Polygon(points, alpha=0.1)
        ax.add_patch(poly)

    ymin_ = ymin*0.1
    ymax_ = ymax*0.1
    ey_ = xr.ecurve.y / xr.ecurve.max_y * ymax_
    points_ = [(0,0)] + list(zip(ex,ey_))+ [(ex[-1],0)]

    for ax in [ax20, ax21,ax22, ax22t]:
        ax.set_ylim(ymin_, ymax_)

    for ax in [ax20, ax21,ax22]:
        poly = Polygon(points_, alpha=0.1)
        ax.add_patch(poly)

    fig.tight_layout()
    plt.show()



"""
    SimpleDecomposeDemo.py

    Copyright (c) 2023-2024, SAXS Team, KEK-PF
"""
import numpy as np
from matplotlib.gridspec import GridSpec
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder
from molass_legacy.KekLib.BasicUtils import Struct

def demo(in_folder, sd_orig, extra_peaks=0, pdf_only=False, modelname="EDM", num_peaks=3):
    from importlib import reload
    import CFSD.SimpleDecompose
    reload(CFSD.SimpleDecompose)    
    from CFSD.SimpleDecompose import decompose_elution_simply
    from molass_legacy.Tools.MapperSingleton import get_mapper

    mapper, sd = get_mapper(sd_orig)
    uv_y = mapper.make_uniformly_scaled_vector(scale=1)

    print(in_folder)
    ecurve = sd.get_xray_curve()
    x = ecurve.x
    y = ecurve.y

    # num_peaks = 3
    if modelname == "EDM":
        import Models.RateTheory.EDM
        reload(Models.RateTheory.EDM)
        from molass_legacy.Models.RateTheory.EDM import EDM, save_reg_data_fh
        model = EDM()
    else:
        import Models.Stochastic.Tripore
        reload(Models.Stochastic.Tripore)
        from molass_legacy.Models.Stochastic.Tripore import Tripore
        save_reg_data_fh = None
        model = Tripore()

    traditional_info = Struct(mapper=mapper, num_peaks=num_peaks)
    decomp_result_non_cfs = decompose_elution_simply(x, y, uv_y, model, traditional_info, using_cfs=False, debug=True)
    if not pdf_only:
        decomp_result_cfs = decompose_elution_simply(x, y, uv_y, model, traditional_info, using_cfs=True)

    if save_reg_data_fh is not None:
        save_reg_data_fh.close()

    def plot_results(ax, x, y, opt_recs):
        ax.plot(x, y)
        cy_list = []
        for rec in opt_recs:
            cy = rec.evaluator(x)
            ax.plot(x, cy, ":", lw=3)
            cy_list.append(cy)
        ty = np.sum(cy_list, axis=0)
        ax.plot(x, ty, ":", color="red", lw=3)

    in_folder_ = get_in_folder(in_folder)
    with plt.Dp():
        fig = plt.figure(figsize=(15,10))
        fig.suptitle("Comparison of %s-decomposition in Different Fitting Spaces with %s" % (model.get_name(), in_folder_), fontsize=20)

        gs = GridSpec(2,5)
        ax10 = fig.add_subplot(gs[0,0])
        ax20 = fig.add_subplot(gs[1,0])
        for ax in ax10, ax20:
            ax.set_axis_off()

        ax10.text(0.5, 0.5, "PDF\nfitting", ha="center", va="center", fontsize=16)
        ax20.text(0.5, 0.5, "CF\nfitting", ha="center", va="center", fontsize=16)

        ax11 = fig.add_subplot(gs[0,1:3])
        ax12 = fig.add_subplot(gs[0,3:5])
        ax21 = fig.add_subplot(gs[1,1:3])
        ax22 = fig.add_subplot(gs[1,3:5])

        ax11.set_title("UV Decomposition", fontsize=16)
        ax12.set_title("XR Decomposition", fontsize=16)

        plot_results(ax11, x, uv_y, decomp_result_non_cfs.opt_recs_uv)
        plot_results(ax12, x, y, decomp_result_non_cfs.opt_recs)

        if not pdf_only:
            plot_results(ax21, x, uv_y, decomp_result_cfs.opt_recs_uv)
            plot_results(ax22, x, y, decomp_result_cfs.opt_recs)

        fig.tight_layout()
        plt.show()



# coding: utf-8
"""

    ファイル名：   RecentFolders.py

    処理内容：

       matplotlib 関連の追加クラス

"""
from datetime       import datetime

"""
    {
        'folder-1'  : [ '2016-03-10 08:44:04.286183', 'folder-1' ],
        'folder-2'  : [ '2016-03-10 08:44:04.395383', 'folder-2' ],
        'folder-3'  : [ '2016-03-10 08:44:04.504583', 'folder-3' ],
    }
"""

class RecentFolders:
    def __init__( self, size=None, settings=None ):
        if settings is None:
            import Settings
            settings = Settings

        if size is None:
            self.size       = settings.get_setting( 'num_recent_folders' )
        else:
            self.size       = size
        self.folder_dict    = settings.get_setting( 'recent_folders' )

    def add( self, path ):
        if len( self.folder_dict ) >= self.size:
            del self.folder_dict[ self.get_sorted_list()[-1][1] ]

        self.folder_dict[ path ] = [ str( datetime.now() ), path ]

    def get_sorted_list( self ):
        list_ = list( reversed( sorted( self.folder_dict.values() ) ) )
        # print( list_ )
        return list_

    def get_current_size( self ):
        return len( self.folder_dict )



"""
    Processes.Recognizer.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
from time import time, sleep

def recognizer(si, pn, queues):
    import molass_legacy.KekLib, SerialAnalyzer
    from molass_legacy._MOLASS.SerialSettings import initialize_settings
    from .SharedInfo import QUEUE_RECOG
    initialize_settings()

    print("recognizer", si)
    recog_queue = queues[QUEUE_RECOG]

    while True:
        i = recog_queue.get()
        if i < 0:
            if i == -1:
                t0 = time()
            elif i == -2:
                print("It took %.3g seconds" % (time()-t0))
            else:
                break
            continue

    print("recognizer stop")



# coding: utf-8
"""

    MemoryUtils.py

    Copyright (c) 2018, Masatsuyo Takahashi, KEK-PF

"""

import sys
from numbers import Number
from collections import Set, Mapping, deque

"""
    borrowed from
    https://stackoverflow.com/questions/449560/how-do-i-determine-the-size-of-an-object-in-python
"""
try: # Python 2
    zero_depth_bases = (basestring, Number, xrange, bytearray)
    iteritems = 'iteritems'
except NameError: # Python 3
    zero_depth_bases = (str, bytes, Number, range, bytearray)
    iteritems = 'items'

def getsize_recursively(obj_0):
    """Recursively iterate to sum size of object & members."""
    def inner(obj, _seen_ids = set()):
        obj_id = id(obj)
        if obj_id in _seen_ids:
            return 0
        _seen_ids.add(obj_id)
        size = sys.getsizeof(obj)
        if isinstance(obj, zero_depth_bases):
            pass # bypass remaining control flow and return
        elif isinstance(obj, (tuple, list, Set, deque)):
            size += sum(inner(i) for i in obj)
        elif isinstance(obj, Mapping) or hasattr(obj, iteritems):
            size += sum(inner(k) + inner(v) for k, v in getattr(obj, iteritems)())
        # Check for custom object instances - may subclass above too
        if hasattr(obj, '__dict__'):
            size += inner(vars(obj))
        if hasattr(obj, '__slots__'): # can have __slots__ with __dict__
            size += sum(inner(getattr(obj, s)) for s in obj.__slots__ if hasattr(obj, s))
        return size
    return inner(obj_0)



# coding: utf-8
"""
    ScatteringCurveModel.py

    Copyright (c) 2018, Masatsuyo Takahashi, KEK-PF
"""
import numpy                as np
import opticspy
import matplotlib.pyplot    as plt

class RapidDebye:
    def __init__( self ):
        pass

    def fit( self, x, y ):
        self.b  = 1
        self.G  = y / ( np.exp( -self.b*q**2 )**2 )

    def eval( self, x ):
        return np.exp( -self.b*x**2 )**2 * self.G



# coding: utf-8
"""

    ProgressChart.py

    Copyright (c) 2017, Masatsuyo Takahashi, KEK-PF

"""

class ProgressChart:
    def __init__( self, fig, size ):
        self.fig    = fig
        self.size   = size
        self.ax1    = fig.add_subplot( 111 )
        self.ax2    = self.ax1.twinx()

    def draw( self, sc_vector,  quality_array, rg_array ):
        self.ax1.cla()
        self.ax2.cla()
        self.ax1.set_axis_off()
        self.ax2.set_axis_off()

        self.ax1.plot( sc_vector, color='blue' )
        self.ax2.plot( rg_array, color='red' )



# coding: utf-8
"""

    ファイル名：   TkSizableWidgets.py

    処理内容：
        Frame( parent, width=..., height=... )
        と同様にサイズ指定できる Widget

        BlinkingLabel   配下のオブジェクトを点滅できる Label

"""
from __future__ import division, print_function, unicode_literals

import os
import sys
if sys.version_info > (3,):
    import tkinter as Tk
    from tkinter import _cnfmerge
else:
    import Tkinter as Tk
    from Tkinter import _cnfmerge

class SizableLabel( Tk.Frame ):
    def __init__( self, parent, width=60, height=26, **kwargs ):
        Tk.Frame.__init__( self, parent, width=width, height=height )
        self.label = Tk.Label( self, **kwargs )
        self.label.pack( fill=Tk.BOTH, expand=Tk.TRUE )

    def pack( self, **kwargs ):
        Tk.Frame.pack( self, **kwargs )
        self.pack_propagate( 0 )        # don't shrink

    def grid( self, **kwargs ):
        Tk.Frame.grid( self, **kwargs )
        self.pack_propagate( 0 )        # don't shrink

    def config( self, **kwargs ):
        state_ = kwargs.get( 'state' )
        if state_ is not None:
            self.label.config( state=state_ )
            del kwargs[ 'state' ]
        Tk.Frame.configure( self, **kwargs )

class SizableEntry( Tk.Frame ):
    def __init__( self, parent, width=60, height=26, **kwargs ):
        Tk.Frame.__init__( self, parent, width=width, height=height )
        self.entry = Tk.Entry( self, **kwargs )
        self.entry.pack( fill=Tk.BOTH, expand=Tk.TRUE )

    def pack( self, **kwargs ):
        Tk.Frame.pack( self, **kwargs )
        self.pack_propagate( 0 )        # don't shrink

    def grid( self, **kwargs ):
        Tk.Frame.grid( self, **kwargs )
        self.pack_propagate( 0 )        # don't shrink

    def config( self, **kwargs ):
        state_ = kwargs.get( 'state' )
        if state_ is not None:
            self.entry.config( state=state_ )
            del kwargs[ 'state' ]
        Tk.Frame.configure( self, **kwargs )

"""
    Thanks to: http://stackoverflow.com/questions/21419032/flashing-tkinter-labels

    revised version of TkSupplements.BlinkingFrame
    TODO: unification of BlinkingLabel and BlinkingFrame
"""
class BlinkingLabel( Tk.Frame ):
    def __init__( self, parent, width=60, height=26, **kwargs ):
        Tk.Frame.__init__( self, parent, width=width, height=height )

        self.label = Tk.Label( self, **kwargs )
        self.label.pack( fill=Tk.BOTH, expand=Tk.TRUE )

        self.switch     = False
        self.reverse    = False

    def pack( self, **kwargs ):
        Tk.Frame.pack( self, **kwargs )
        self.pack_propagate( 0 )        # don't shrink

    def grid( self, **kwargs ):
        Tk.Frame.grid( self, **kwargs )
        self.pack_propagate( 0 )        # don't shrink

    def flash(self):
        if not self.reverse and not self.switch:
            return

        try:
            bg = self.label.cget("background")
            fg = self.label.cget("foreground")
            self.label.configure(background=fg, foreground=bg)
        except:
            pass

        self.reverse = not self.reverse
        self.after(250, self.flash)

    def start(self):
        if not self.switch:
            self.switch = True
            self.flash()

    def stop(self):
        if self.switch:
            self.switch = False

    def is_blinking( self ):
        return self.switch



"""
    ElutionMatrix.py - successor to the LRF.ConcMatrix

    refactoring phase 1     using ConcMatrix
    refactoring phase 2     rewrite

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import logging
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt
from .ConcMatrix import ConcMatrix

USE_DICTINCT_CD = True

class ElutionMatrix:
    def __init__(self, depr_solver):
        self.logger = logging.getLogger(__name__)
        self.conc = depr_solver.conc
        self.conc_factor = depr_solver.conc_factor
        self.ecurve = depr_solver.ecurve
        self.j0 = depr_solver.j0
        self.data = depr_solver.data
        self.cd_slice = depr_solver.cd_slice
        self.xray_scale = depr_solver.xray_scale
        self.mapper = depr_solver.mapper
        self.conc_depend = depr_solver.conc_depend
        self.paired_ranges = depr_solver.cnv_ranges
        self.rank_control = depr_solver.rank_control
        self.mc_vector = depr_solver.mc_vector

    def make_elution_matrix(self, start, stop, peakset_info, lrf_rank=None, debug=False):
        pno, nth, peakset, known_peak_info = peakset_info
        paired_ranges_ = [self.paired_ranges[i] for i in peakset]

        x = self.ecurve.x[start:stop]
        if self.mc_vector is None:
            mc_vector = None
        else:
            mc_vector = self.mc_vector[start:stop]

        if self.rank_control:
            if lrf_rank is None:
                if USE_DICTINCT_CD:
                    cmatrix = ConcMatrix(x, self.conc, conc_depend=1,
                                            paired_ranges=paired_ranges_, mc_vector=mc_vector,
                                            conc_factor=self.conc_factor,
                                            ecurve=self.ecurve, j0=self.j0,
                                            )
                    M = self.data[:,start:stop]
                    conc_depend = compute_distinct_cd(M, cmatrix.data, self.cd_slice, self.xray_scale, self.logger)
                else:
                    conc_depend = self.mapper.get_cd_degree_from_range(start, stop)
            else:
                if lrf_rank == 1:
                    conc_depend = 1
                else:
                    conc_depend = 2
        else:
            conc_depend = self.conc_depend

        cmatrix = ConcMatrix(x, self.conc, conc_depend=conc_depend,
                                paired_ranges=paired_ranges_, mc_vector=mc_vector,
                                conc_factor=self.conc_factor,
                                ecurve=self.ecurve, j0=self.j0,
                                )

        if debug:
            x_ = self.ecurve.x
            y_ = self.ecurve.y
            with plt.Dp():
                fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
                fig.suptitle("make_elution_matrix: [%d:%d]" % (start, stop))
                if self.mc_vector is not None:
                    ax1.plot(x_, self.mc_vector)
                for k, cy in enumerate(cmatrix.data):
                    if k%2 == 0:
                        ax1.plot(x, cy, ":", lw=3)
                ax2.plot(x_, y_)
                fig.tight_layout()
                plt.show()

        return cmatrix.data



"""
    OurTkinter.py

    Copyright (c) 2016-2024, Masatsuyo Takahashi, KEK-PF
"""
import os
import gc

import OurTk as Tk
from tkinter import TkVersion
# from tkinter.simpledialog import Dialog
from tkinter import filedialog as FileDialog
from tkinter.scrolledtext import ScrolledText
import tkinter.font as Font
import tkinter.ttk as ttk
# import tkinter.tix as tix

from molass_legacy.KekLib.TkSupplements      import tk_set_icon_portable     # for backward compatibility

try:
    from idlelib.tooltip import Hovertip as ToolTip
except:
    from idlelib.ToolTip import ToolTip

# from molass_legacy.KekLib.TkSupplements      import tk_set_icon_portable

def is_empty_val( val ):
    return val is None or val == '' or val[0] == '<' or val == 'None' 

def checkFolder( parent, folder_var, folder_entry ):
    # TODO: include these checks in a custom widget class
    def set_error( entry ):
        entry.config( fg='red' )
        entry.focus_force()

    ok_ = False
    folder = folder_var.get()
    if is_empty_val( folder ):
        set_error( folder_entry )
        MessageBox.showerror( "Input Folder Error", "The Input Folder is required.", parent=parent )
    elif not os.path.exists( folder ):
        set_error( folder_entry )
        MessageBox.showerror( "Input Folder Error", "'%s' does not exist." % in_folder, parent=parent )
    elif not os.path.isdir( folder ):
        set_error( folder_entry )
        MessageBox.showerror( "Input Folder Error", "'%s' is not a folder." % in_folder, parent=parent )
    else:
        folder_entry.config( fg='black' )
        ok_ = True

    return ok_


class Dialog(Tk.Toplevel):

    '''Class to open dialogs.

    This class is intended as a base class for custom dialogs
    '''

    def __init__(self, parent, title=None, auto_geometry=True, geometry_cb=None, visible=True, block=True, location=None ):

        '''Initialize a dialog.

        Arguments:

            parent -- a parent window (the application window)

            title -- the dialog title
        '''
        Tk.Toplevel.__init__(self, parent)

        self.withdraw() # remain invisible for now
        # If the master is not viewable, don't
        # make the child transient, or else it
        # would be opened withdrawn
        if parent.winfo_viewable():
            self.transient(parent)

        if title:
            self.title(title)

        self.parent = parent
        self.location = location

        self.result = None

        body = Tk.Frame(self)
        self.initial_focus = self.body(body)
        body.pack(padx=5, pady=5)

        self.buttonbox()

        if not self.initial_focus:
            self.initial_focus = self

        self.protocol("WM_DELETE_WINDOW", self.cancel)

        if auto_geometry:
            self._adjust_geometry()

        if geometry_cb is not None:
            geometry_cb()

        if not visible:
            return

        self._show(block=block)

    def _show(self, block=True, pause=None, wait=True, raise_=False):
        self.deiconify() # become visible now

        self.initial_focus.focus_set()

        # wait for window to appear on screen before calling grab_set
        try:
            self.wait_visibility()
        except:
            if raise_:
                raise Exception()
            """
            ignore because this does not seem to be a problem
            _tkinter.TclError: window "..." was deleted before its visibility changed
            """
            pass

        try:
            self.grab_set()
        except:
            if raise_:
                raise Exception()
            """
            ignore because this does not seem to be a problem
            _tkinter.TclError: grab failed: window "..." already has grab
            """
            pass

        if not block:
            interval = 500 if pause is None else int(pause*1000)
            self.parent.after( interval, self.ok )

        if wait:
            self.wait_window(self)
        else:
            # user should self.wait_window instead
            pass

    def destroy(self):
        '''Destroy the window'''
        self.initial_focus = None
        Tk.Toplevel.destroy(self)

        # to avoid "Tcl_AsyncDelete: async handler deleted by the wrong thread" error
        #   from https://pysimplegui.readthedocs.io/en/latest/#multiple-threads
        gc.collect()

    #
    # construction hooks

    def body(self, master):
        '''create dialog body.

        return widget that should have initial focus.
        This method should be overridden, and is called
        by the __init__ method.
        '''
        pass

    def buttonbox( self, frame=None ):
        '''add standard button box.

        override if you do not want the standard buttons
        '''

        if frame is None:
            box = Tk.Frame(self)
            box.pack()
        else:
            box = frame

        w = Tk.Button(box, text="OK", width=10, command=self.ok, default=Tk.ACTIVE)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        w = Tk.Button(box, text="Cancel", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=5, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    #
    # standard button semantics

    def ok(self, event=None):

        if not self.validate():
            self.initial_focus.focus_set() # put focus back
            return

        self.withdraw()
        self.update_idletasks()

        try:
            self.apply()
        finally:
            self.cancel()

    def cancel(self, event=None):

        # put focus back to the parent window
        if self.parent is not None:
            self.parent.focus_set()
        self.destroy()

    #
    # command hooks

    def validate(self):
        '''validate the data

        This method is called automatically to validate the data before the
        dialog is destroyed. By default, it always validates OK.
        '''

        return 1 # override

    def apply(self):
        '''process the data

        This method is called automatically to process the data, *after*
        the dialog is destroyed. By default, it does nothing.
        '''

        pass # override

    def _adjust_geometry(self):
        """
        use this method with Tk-after method before _show
            when
            used with visible=False
            and the window size is not yet what is expected
                during the above __init__
        """

        parent = self.parent
        location = self.location

        rootx = parent.winfo_rootx()
        rooty = parent.winfo_rooty()
        if location is None:
            self.geometry("+%d+%d" % (rootx+50,rooty+30))
        else:
            vloc, hloc = location.split(" ")
            self.update()
            from molass_legacy.KekLib.TkUtils import split_geometry
            W, H, X, Y = split_geometry(parent.geometry())
            w, h, x, y = split_geometry(self.geometry())
            # print([rootx, rooty])
            # print([W, H, X, Y], [w, h, x, y])
            if vloc == 'lower':
                offsety = H - h - 100
            elif vloc == 'center':
                offsety = H//2 - h//2
            else:
                offsety = 30

            if hloc == 'right':
                offsetx = W - w - 100
                self.geometry("+%d+%d" % (rootx+offsetx, rooty+offsety))
            elif hloc == 'center':
                offsetx = W//2 - w//2
            else:
                offsetx = 50
 
            self.geometry("+%d+%d" % (rootx+offsetx, rooty+offsety))



import numpy                as np

#   svergun-1987 (1.27)
def f( sR ):
     return 3 * ( np.sin( sR ) - sR * np.cos( sR ) ) / ( sR )**3

class Intensity:
    def __init__( self, N, R ):
        self.N  = N
        self.R  = R
        self.V  = 4 * np.pi * R**3 / 3
        self.v1 = self.V / N        # volume per particle

    def term1( self, x ):
        sR  = x * self.R
        return self.N * f(sR)**2

    def term2( self, x ):
        sR  = x * self.R
        return 8 * self.V / self.v1 * f( 2*sR )

    def __call__( self, x ):
        return self.term1( x ) * ( 1 - self.term2( x ) )



# coding: utf-8
"""
    CleanerThread.py

    craetes a thread which terminates its own process
    when its parent's process have died

    borrowed from
    Python: how to kill child process(es) when parent dies?
    https://stackoverflow.com/questions/23434842/python-how-to-kill-child-processes-when-parent-dies
"""
import os
import logging
from ctypes import WinDLL, WinError
from ctypes.wintypes import DWORD, BOOL, HANDLE
from molass_legacy.KekLib.KillableThread import Thread

class CleanerThread:
    def __init__(self, parent_pid, cleanup=None):
        self.parent_pid = parent_pid
        self.cleanup = cleanup
        self._start()

    def start(self):
        pass

    def _start(self):
        self.thread = Thread( target=self.wait_proc, name='WaitThread', args=[] )
        self.thread.start()

    def wait_proc(self):
        # Magic value from http://msdn.microsoft.com/en-us/library/ms684880.aspx
        SYNCHRONIZE = 0x00100000
        kernel32 = WinDLL("kernel32.dll")
        kernel32.OpenProcess.argtypes = (DWORD, BOOL, DWORD)
        kernel32.OpenProcess.restype = HANDLE
        parent_handle = kernel32.OpenProcess(SYNCHRONIZE, False, self.parent_pid)
        # Block until parent exits
        try:
            os.waitpid(parent_handle, 0)
        except Exception as exc:
            print(exc)
            # [Errno 13] Permission denied
        logger = logging.getLogger(__name__)

        if self.cleanup is not None:
            logger.info("child process cleanup call.")
            self.cleanup()
            logger.info("child process cleanup done.")

        pid = os.getpid()
        logger.info("terminates child process %d of died parent process %d", pid, self.parent_pid)
        os._exit(0)

    def set_cleanup(self, cleanup):
        self.cleanup = cleanup

    def __del__(self):
        self.thread.terminate()



"""
    CfsEvalPeaksDemo.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize
from matplotlib.gridspec import GridSpec
from molass_legacy.KekLib.SciPyCookbook import smooth
from DataUtils import get_in_folder

VERY_SMALL_VALUE = 1e-6

def demo(in_folder, sd, extra_peaks=0):
    import molass_legacy.KekLib.DebugPlot as plt
    from molass_legacy.Models.ElutionCurveModels import egha, emga
    from SecTheory.Edm import edm_func
    from importlib import reload
    import Models.CfsEvalPeaks
    reload(Models.CfsEvalPeaks)
    from .CfsEvalPeaks import get_cfs_evaluated_peaks

    print(in_folder)
    ecurve = sd.get_xray_curve()
    x = ecurve.x
    y = ecurve.y

    num_peaks = len(ecurve.peak_info) + extra_peaks
    egh_peak_list, emg_peak_list, edm_peak_list = get_cfs_evaluated_peaks(x, y, num_peaks, using_cfs=False, debug=True)
    print("egh_peak_list=", egh_peak_list)
    print("emg_peak_list=", emg_peak_list)

    in_folder_ = get_in_folder(in_folder)
    with plt.Dp():
        fig = plt.figure(figsize=(20,12))
        fig.suptitle("Fitting Difference Demo between PDF and CF using %s" % in_folder_, fontsize=24)

        gs = GridSpec(3,15)
        ax0 = fig.add_subplot(gs[0,0])

        ax1 = fig.add_subplot(gs[0,1:3])
        ax2 = fig.add_subplot(gs[0,3:7])
        ax3 = fig.add_subplot(gs[0,7:11])
        ax4 = fig.add_subplot(gs[0,11:15])

        ax5 = fig.add_subplot(gs[1,0])
        ax6 = fig.add_subplot(gs[1,1:3])
        ax7 = fig.add_subplot(gs[1,3:7])
        ax8 = fig.add_subplot(gs[1,7:11])
        ax9 = fig.add_subplot(gs[1,11:15])

        axa = fig.add_subplot(gs[2,0])
        axb = fig.add_subplot(gs[2,1:3])
        axc = fig.add_subplot(gs[2,3:7])
        axd = fig.add_subplot(gs[2,7:11])
        axe = fig.add_subplot(gs[2,11:15])

        ax0.set_title("No", fontsize=16)
        ax1.set_title("Space", fontsize=16)
        ax2.set_title("EGHA", fontsize=16)
        ax3.set_title("EMGA", fontsize=16)
        ax4.set_title("EDM", fontsize=16)

        def plot_params(ax0, ax1, ax2, ax3, ax4, text0, text1, egh_peak_list, emg_peak_list, edm_peak_list, proportional=False):

            for ax, text, fontsize in (ax0, text0, 50), (ax1, text1, 16):
                ax.grid(False)
                ax.set_xticklabels([])
                ax.set_yticklabels([])
                ax.set_facecolor("white")
                ax.text(0.5, 0.5, text, ha="center", va="center", fontsize=fontsize)

            for func, ax, peak_list in [
                            (egha, ax2, egh_peak_list),
                            (emga, ax3, emg_peak_list),
                            (edm_func, ax4, edm_peak_list),
                            ]:
                ax.plot(x, y)

                cy_list = []
                for p in peak_list:
                    cy = func(x, *p)
                    cy_list.append(cy)
                ty = np.sum(cy_list, axis=0)

                if proportional:
                    sy = smooth(y)
                    cy_list_ = []
                    ty[ty < VERY_SMALL_VALUE] = VERY_SMALL_VALUE
                    ty_ = np.zeros(len(x))
                    for cy in cy_list:
                        cy_ = sy * cy/ty
                        cy_list_.append(cy_)
                        ty_ += cy_
                else:
                    ty_ = ty
                    cy_list_ = cy_list

                for cy in cy_list_:
                    ax.plot(x, cy, ":")

                ax.plot(x, ty_, ":", color="red", lw=3)

        plot_params(ax0, ax1, ax2, ax3, ax4, "1", "Elution Space\nevaluated", egh_peak_list, emg_peak_list, edm_peak_list)

        egh_peak_list, emg_peak_list, edm_peak_list = get_cfs_evaluated_peaks(x, y, num_peaks, using_cfs=True, debug=True)
        plot_params(ax5, ax6, ax7, ax8, ax9, "2", "CF Space\nevaluated", egh_peak_list, emg_peak_list, edm_peak_list)
        plot_params(axa, axb, axc, axd, axe, "3", "CF Space\nevaluated\n\nwith\n\nProportional\nAllocation", egh_peak_list, emg_peak_list, edm_peak_list, proportional=True)

        fig.tight_layout()
        fig.subplots_adjust(top=0.88, left=0.05,  bottom=0.08, right=0.95)
        plt.show()



"""
    Optimizer.FvScoreInspecor.py

    Copyright (c) 2023-2024, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.widgets import Slider
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
from molass_legacy.KekLib.ScrolledFrame import ScrolledFrame
from molass_legacy._MOLASS.SerialSettings import get_setting

class FvScoreInspecor(Dialog):
    def __init__(self, parent, js_canvas):
        self.parent = parent
        self.js_canvas = js_canvas
        self.dsets = dsets = js_canvas.dsets
        self.optimizer = js_canvas.fullopt
        self.score_names = self.optimizer.get_score_names(major_only=True)
        self.x_array = js_canvas.demo_info[1]
        Dialog.__init__(self, parent, "FV Score Inspection", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):
        cframe = ScrolledFrame(body_frame)
        cframe.pack(side=Tk.LEFT)
        cframe.add_bind_mousewheel()

        cframe_ = Tk.Frame(cframe.interior)
        cframe_.pack()
        tframe_ = Tk.Frame(cframe.interior)
        tframe_.pack(side=Tk.LEFT)

        self.compute_scores()

        fig = plt.figure(figsize=(18,11))
        self.fig = fig
        self.draw_scores()
        self.mpl_canvas = FigureCanvasTkAgg(fig, cframe_)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack( fill=Tk.BOTH, expand=1)
        self.toolbar = NavigationToolbar(self.mpl_canvas, tframe_)
        self.toolbar.update()
        fig.tight_layout()
        self.mpl_canvas.draw()

    def compute_scores(self):
        fv_variations = []
        # NUM_MAJOR_SCORES = get_setting("NUM_MAJOR_SCORES")
        # self.score_variations = np.zeros((self.x_array.shape[0], NUM_MAJOR_SCORES))
        self.score_variations = np.zeros((self.x_array.shape[0], 13))
        for i, p in enumerate(self.x_array):
            p = self.x_array[i,:]
            ret = self.optimizer.objective_func(p, return_full=True)
            if np.isscalar(ret):
                fv = ret
                score_list = np.ones(6)*np.nan
            else:
                fv, score_list = ret[0:2]
            print([i], fv, score_list)
            fv_variations.append(fv)
            self.score_variations[i,:] = score_list

        self.fv_variations = np.array(fv_variations)

    def draw_scores(self):
        fig = self.fig
        ax = fig.add_subplot(111)
        ax.set_title("Variation of FV Scores", fontsize=20)

        for i, label in enumerate(self.score_names):
            ax.plot(self.score_variations[:,i], ":", label=label)

        ax.plot(self.fv_variations, color="red", label="FV")

        ax.set_ylim(-2, 2.5)

        ax.legend()



"""

    ファイル名：    LogDiff.py

    処理内容：      ログファイルの diff

    Copyright (c) 2017,2023, Masatsuyo Takahashi, KEK-PF

"""
import re
from difflib    import unified_diff
from io         import StringIO

log_diff_re = re.compile( r'^\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}' )
time_re     = re.compile( r'was \d\.\d+ seconds' )

def time_re_sub( line ):
    return re.sub( time_re, 'was S.SSS seconds', line )

def make_log_diff(from_file, to_file, re_sub=None, write_substituted=False):

    def substitute( line ):
        line = re.sub( log_diff_re, 'YYYY-mm-dd HH:MM:SS', line )
        if re_sub is not None:
            line = re_sub( line )
        return line

    log_text_list = []
    for file in [from_file, to_file]:
        fh = open( file )
        text = [ substitute(line) for line in fh ]
        if write_substituted:
            path = file.replace(".log", ".txt")
            with open(path, "w") as out_fh:
                out_fh.write("".join(text))
        log_text_list.append( text )

    buff = StringIO()
    buff.writelines( unified_diff( log_text_list[0], log_text_list[1], fromfile=from_file, tofile=to_file ) )

    return buff.getvalue()

if __name__ == '__main__':
    import sys
    print(sys.argv)
    assert len(sys.argv) == 3
    print(make_log_diff(*sys.argv[1:3], write_substituted=True))



"""

    ファイル名：   PilatusCounter.py

    処理内容：

       Pilatus 画像データ関連の共通処理

    Copyright (c) 2015-2024, Masatsuyo Takahashi, KEK-PF
"""
from __future__ import division, print_function, unicode_literals
    # /   を使うと商は float
    # //  を使うと商は int

import os
import sys
import glob
import re
import pandas as pd
from collections import defaultdict

class Counter:
    def __init__( self, in_folder ):
        self.in_folder = in_folder
        self.counter_tables = []

        orig_folder = os.getcwd()
        os.chdir( self.in_folder )
        counter_files   = glob.glob( 'PilatusCounter_*.txt' )

        # print( " counter_files=", counter_files )

        counter_dict = {}
        for file in counter_files:
            counter_table = pd.read_table( file )
            self.counter_tables.append( counter_table )

        os.chdir( orig_folder )

    def get_counter_dict( self, counter_id ):
        counter_dict = {}

        for counter_table in self.counter_tables:
            if counter_id == 'None':
                counter = None
            else:
                counter = counter_table[ counter_id ]       # counter_id の列を選ぶ。
            for i in range( counter_table.File.size ):
                # 'test001_0_d0_00000.tif' などの場合、末尾の拡張子を削除する。
                file_id = re.sub( r'\.\w+$', '', counter_table.File[i] )

                # 'MAG2wk7210_0_00000-0' などの '_00000-0' または '_00000' を削除する。
                # なお、枝番（-0）はある時期に一時的に存在したもので、
                # 最新の状態では存在しないものである。
                # fkey = re.sub( r'_\d+(-\d+)?$', '', file_id )

                fkey = file_id

                if counter_id == 'None':
                    count = 1
                else:
                    count = counter[i]
                counter_dict[ fkey ] = count

        return counter_dict

    def available_counters( self ):
        available_keys = defaultdict( lambda: 0 )
        for counter_table in self.counter_tables:
            for colname, series in counter_table.iteritems():
                if colname[0] != 'C':
                    continue
                # print colname, series
                available = True
                for num in series:
                    if num <= 0:
                        available = False
                        break
                if available:
                    available_keys[ colname ] += 1

        return sorted( available_keys.keys() )



# coding: utf-8
"""
    Baseline.LinearAdjustment.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import logging
import numpy as np
from molass_legacy._MOLASS.SerialSettings import set_setting

LA_MIN_SCI = 80

class LaJudge:
    def __init__(self, mapper):
        self.logger = logging.getLogger(__name__)
        self.mapper = mapper
        self.min_sci = np.min(mapper.get_sci_list())
        if False:
            from molass_legacy.Mapping.MappingUtils import debug_plot_mapping
            debug_plot_mapping(mapper)

    def ok(self):
        return self.min_sci >= LA_MIN_SCI

    def modify_opt_params(self):
        if not self.ok():
            opt_params = self.mapper.opt_params
            if opt_params['uv_baseline_adjust']:
                opt_params['uv_baseline_adjust'] = 0
                self.logger.info("UV baseline adjustment has been suppressed due to a low SCI %d.", int(self.min_sci))
                set_setting('uv_adjust_suppressed', 1)
                self.mapper.optimize(opt_params=opt_params)



"""
    StateSequence.py

    Copyright (c) 2022-2024, SAXS Team, KEK-PF
"""
import os
import logging
import numpy as np
from datetime import datetime

NUM_FUNC_CALLS_LOCALOPT = 5000

class StateSequence:
    def __init__(self, work_folder=None, niter=100, seed=None):
        if work_folder is None:
            self.fv_array = np.array([])
            self.x_array = np.array([])
            self.xmax = niter * NUM_FUNC_CALLS_LOCALOPT
            if seed is None:
                seed = np.random.randint(100000, 999999)
            self.seed = seed
        else:
            self.fv_array, self.x_array, self.xmax, self.seed = self.get_info_from_folder(work_folder)

    def get_info(self):
        return self.fv_array, self.x_array, self.xmax, self.seed

    def get_nc(self):
        return self.logfile.nc

    def insert(self, fv, x):
        self.fv_array = np.concatenate([self.fv_array, (None, fv, None, None)])     # [fv] must be consistent with (counter, fv, accept, time)
        self.x_array = np.concatenate([self.x_array, [x]])

    def __repr__(self):
        return "StateSequence(%d, %d, %s, %s)" % (len(self.fv_array), len(self.x_array), str(self.xmax), str(self.seed))

    def get_info_from_folder(self, work_folder, demo_index=0):
        from .OptLogFile import OptLogFile
        self.logfile = OptLogFile(os.path.join(work_folder, "optimizer.log"))
        seed_txt = os.path.join(work_folder, 'seed.txt')
        seed = read_seed_txt(seed_txt)
        cb_file = os.path.join(work_folder, 'callback.txt')
        fv_list, x_list = read_callback_txt_impl(cb_file)
        if demo_index > 0:
            n = demo_index + 1
        else:
            n = len(fv_list)
        return np.array(fv_list[0:n]), np.array(x_list[0:n]), fv_list[-1][0], seed

    def get_in_folder(self):
        return self.logfile.optdict['-f']

def save_opt_params(fh, x, f, accept, eval_counter):
    precision_save = np.get_printoptions()["precision"]
    np.set_printoptions(precision=17)

    fh.write("t=%s\n" % datetime.now().strftime("%Y-%m-%d %H:%M:%S"))
    fh.write("x=\n%s\n" % str(x))
    fh.write("f=%g\n" % f)
    fh.write("a=%s\n" % str(accept))
    fh.write("c=%d\n" % eval_counter)
    fh.flush()

    np.set_printoptions(precision=precision_save)

def read_callback_txt_impl(cb_file):
    import os
    import re
    from datetime import datetime
    from molass_legacy.KekLib.NumpyArrayUtils import from_space_separated_list_string

    time_re = re.compile(r't=(\d{4})-(\d{2})-(\d{2}) (\d{2}):(\d{2}):(\d{2})')
    space_re = re.compile(r'\s+')
    fv_re = re.compile(r'f=(\S+)')   # f=-0.375527 as well as f=138548, f=-8.52599e-05
    accept_re = re.compile(r'a=(\w+)')
    counter_re = re.compile(r'c=(\d+)')

    fv_list = []
    x_list = []
    accept = None
    time = None

    if os.path.exists(cb_file):
        fh = open(cb_file)
        for k, line in enumerate(fh):
            # print([k], line)
            if line[0:2] == "t=":
                m = time_re.search(line)
                if m:
                    year = int(m.group(1))
                    month = int(m.group(2))
                    day = int(m.group(3))
                    hour = int(m.group(4))
                    minute = int(m.group(5))
                    second = int(m.group(6))
                    time = datetime(year, month, day, hour, minute, second)
            elif line[0:2] == "x=":
                pass
            elif line[0] == "[":
                x_str = line
            elif line[0] == " ":
                x_str += line
            elif line[0:2] == "f=":
                m = fv_re.search(line)
                if m:
                    fv = float(m.group(1))
            elif line[0:2] == "a=":
                m = accept_re.search(line)
                if m:
                    accept = m.group(1) == 'True'
            elif line[0:2] == "c=":
                m = counter_re.search(line)
                if m:
                    counter = int(m.group(1))
                    fv_list.append((counter, fv, accept, time))
                    x = from_space_separated_list_string(x_str)
                    x_list.append(x)
        fh.close()
    else:
        # not yet written
        pass

    return fv_list, x_list

def read_seed_txt(seed_txt):
    import re

    seed_re = re.compile(r'seed=(\d+)')
    try:
        with open(seed_txt) as fh:
            for line in fh:
                m = seed_re.search(line)
                if m:
                    seed = int(m.group(1))
    except:
        seed = None
    return seed



"""
    Particles.SmcModel.py

    Copyright (c) 2024, SAXS Team, KEK-PF    
"""
from particles import smc_samplers as ssp

class SmcModel(ssp.StaticModel):
    def __init__(self, optimizer):
        self.optimizer = optimizer

    def logpyt(self, theta, t):  # density of Y_t given theta and Y_{0:t-1}
        # 
        return stats.norm.logpdf(self.data[t], loc=theta['mu'],
                                 scale = theta['sigma']) 


"""
    Selective.V1ParamsAdapter.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Decomposer.ModelEvaluator import ModelEvaluator
from molass_legacy.Decomposer.FitRecord import FitRecord
from molass_legacy.Decomposer.UnifiedDecompResult import UnifiedDecompResult
from .PeakProxy import PeakProxy

def make_decomp_result_impl(editor, peaks, debug=False):
    cf = editor.get_current_frame()
    x = cf.x
    y = cf.y
    fx = cf.fx
    model = cf.model

    x_curve = editor.corbase_info.x_curve
    max_y = x_curve.max_y
    uv_y = cf.uv_y
    max_y_uv = np.max(uv_y)

    opt_recs = make_xr_opt_recs_adapted(model, fx, y, peaks)
    uv_scale = max_y_uv/max_y
    opt_recs_uv = make_uv_opt_recs_adapted(model, fx, uv_y, peaks, uv_scale)

    if debug:
        from importlib import reload
        import Decomposer.OptRecsUtils
        reload(Decomposer.OptRecsUtils)
        from molass_legacy.Decomposer.OptRecsUtils import debug_plot_opt_recs_impl
        from molass_legacy.Elution.CurveUtils import simple_plot
        with plt.Dp():
            fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
            fig.suptitle("make_decomp_result_impl debug")
            ax1.plot(fx, uv_y, color="blue")
            debug_plot_opt_recs_impl(ax1, fx, uv_y, opt_recs_uv, color="blue")
            ax2.plot(fx, y, color="orange")
            debug_plot_opt_recs_impl(ax2, fx, y, opt_recs, color="orange")
            fig.tight_layout()
            plt.show()

    decomp_result =  UnifiedDecompResult(
                xray_to_uv=None,
                x_curve=x_curve, x=x_curve.x, y=x_curve.y,
                opt_recs=opt_recs,
                max_y_xray = max_y,
                model_name=model.get_name(),
                decomposer=None,
                uv_y=uv_y,
                opt_recs_uv=opt_recs_uv,
                max_y_uv = max_y_uv,
                debug_info=None,
                )

    decomp_result.set_area_proportions()
    decomp_result.remove_unwanted_elements()    # required to compute proportions used in decomp_result.identify_ignorable_elements()
    return decomp_result

def make_xr_opt_recs_adapted(model, fx, y, peaks):
    chisqr_n = np.nan
    ret_recs = []
    top_y_list = []
    for kno, params in enumerate(peaks):
        evaluator = ModelEvaluator(model, params, sign=1)
        y_ = evaluator(fx)
        m = np.argmax(y_)
        top_y = y_[m]
        top_y_list.append(top_y)
        peak = PeakProxy(top_x=fx[m], top_y=top_y)
        fit_rec = FitRecord(kno, evaluator, chisqr_n, peak)
        ret_recs.append(fit_rec)
    max_y = np.max(top_y_list)
    for kno, fit_rec in enumerate(ret_recs):
        fit_rec.peak.top_y_ratio = fit_rec.peak.top_y/max_y
    return ret_recs

def make_uv_opt_recs_adapted(model, fx, uv_y, peaks, scale):
    if model.is_traditional():
        converted_list = []
        for kno, params in enumerate(peaks):
            params_ = params.copy()
            params_[0] *= scale         # this won't work for EDM which is not traditional
            converted_list.append(params_)
    else:
        # note that non traditional models must implement this method
        converted_list = model.adjust_to_xy(peaks, fx, uv_y)

    chisqr_n = np.nan            
    ret_recs = []
    top_y_list = []
    for kno, params in enumerate(converted_list):
        evaluator = ModelEvaluator(model, params, sign=1)
        y_ = evaluator(fx)
        m = np.argmax(y_)
        top_y = y_[m]
        top_y_list.append(top_y)
        peak = PeakProxy(top_x=fx[m], top_y=top_y)
        fit_rec = FitRecord(kno, evaluator, chisqr_n, peak)
        ret_recs.append(fit_rec)
    max_y = np.max(top_y_list)
    for kno, fit_rec in enumerate(ret_recs):
        fit_rec.peak.top_y_ratio = fit_rec.peak.top_y/max_y
    return ret_recs


# coding: utf-8
"""

    ファイル名：   TkSupplements.py

    処理内容：

        BlinkingFrame   配下のオブジェクトを点滅できる Frame

    Copyright (c) 2016-2021, Masatsuyo Takahashi, KEK-PF

"""
from __future__ import division, print_function, unicode_literals

import os
import sys
import inspect
if sys.version_info > (3,):
    import tkinter as Tk
    from tkinter import _cnfmerge
else:
    import Tkinter as Tk
    from Tkinter import _cnfmerge

def pad_text( text, length ):
    return text + ' ' * ( length - len(text) )

from molass_legacy.KekLib.BasicUtils import exe_name, get_caller_module

"""
    Icon files are supposed to be placed in the same folder,
    named as follows.
        name.ico on Windows
        name.png on Linux
"""

cashed_image = None

def tk_set_icon_portable( parent, name=None, module=None, use_cashed_image=False ):
    pass

def set_icon( parent ):
    iconpath = os.environ.get('OUR_ICON_PATH')
    if iconpath is None:
        return

    if os.name == 'nt':         # i.e. on Windows
        if not os.path.exists(iconpath): return
        # see wm_iconbitmap in tkinter.__init__.py
        # see also https://wiki.tcl-lang.org/page/wm+iconbitmap
        parent.iconbitmap(default=iconpath)
    else:
        if not os.path.exists(iconpath): return
        img = Tk.Image( "photo", file=iconpath )
        parent.tk.call( 'wm', 'iconphoto', parent._w, img )

"""
    Thanks to: http://stackoverflow.com/questions/21419032/flashing-tkinter-labels
"""
FLASH_STOP_NORMAL   = 0
FLASH_STOP_JUST     = 1

class BlinkingFrame(Tk.Frame):
    def __init__( self, parent, object_spec_array=None, grid=False,
                    start_proc=None, stop_proc=None,
                    flash_stop_type=FLASH_STOP_NORMAL,
                    debug=False ):
        Tk.Frame.__init__(self, parent)

        self.objects = []

        if object_spec_array is not None:
            for object_spec in object_spec_array:
                obj, pack_info  = object_spec
                self.objects.append( obj )
                if grid:
                    obj.grid( **pack_info )
                else:
                    obj.pack( **pack_info )

        # print('self.objects=', self.objects)

        self.start_proc = start_proc
        self.stop_proc  = stop_proc
        self.switch     = False
        self.reverse    = False
        self.debug      = debug
        self.flash_stop_type    = flash_stop_type
        self.init_bgfg = None

    def flash(self):
        if self.flash_stop_type == FLASH_STOP_NORMAL:
            return_cond = ( not self.reverse and not self.switch )
        else:
            return_cond = not self.switch

        if return_cond:
            return

        for obj in self.objects:
            try:
                bg = obj.cget("background")
                fg = obj.cget("foreground")
                obj.configure(background=fg, foreground=bg)
            except:
                continue

        self.reverse = not self.reverse
        self.after(250, self.flash)

    def start(self):
        if self.init_bgfg is None:
            self.init_bgfg = [ (obj.cget("background"), obj.cget("foreground")) for obj in self.objects ]
        if not self.switch:
            if self.debug:
                print( 'start blinking' )
            self.switch = True
            if self.start_proc:
                self.start_proc()
            self.flash()

    def stop(self):
        self.restore_init()
        if self.switch:
            if self.debug:
                print( 'stop blinking' )
            self.switch = False
            if self.stop_proc:
                self.stop_proc()

    def restore_init(self):
        if self.init_bgfg is not None:
            for k, obj in enumerate(self.objects):
                bg, fg = self.init_bgfg[k]
                obj.configure(background=bg, foreground=fg)
        self.reverse = False

    def is_blinking( self ):
        return self.switch

"""
    pixel-sizable button widget
    # cf. http://www-acc.kek.jp/kekb/control/Activity/Python/TkIntro/introduction/button.htm
    TODO: doesn't seem to work when placed by grid
"""
class SlimButton(Tk.Frame):
    def __init__( self, parent, text, command, state='normal', height=26, width=60 ):
        Tk.Frame.__init__( self, parent, height=height, width=width )
        self.button = Tk.Button( self, text=text, command=command, state=state )
        self.button.pack( fill=Tk.BOTH, expand=1 )

    def pack( self, cnf={}, **kw ):
        cnf = _cnfmerge((cnf, kw))      # cf. tkinter.dialog.py
        Tk.Frame.pack( self, cnf )
        self.pack_propagate( 0 )        # don't shrink

    def invoke( self ):
        # TODO: other args?
        self.button.invoke()

    def configure( self, cnf={}, **kw ):
        cnf = _cnfmerge((cnf, kw))      # cf. tkinter.dialog.py
        self.button.configure( cnf )

    def config( self, cnf={}, **kw ):
        cnf = _cnfmerge((cnf, kw))      # cf. tkinter.dialog.py
        self.button.configure( cnf )

    def cget( self, *args, **kw ):
        return self.button.cget( *args, **kw )

if __name__ == "__main__":
    root = Tk.Tk()

    objects = []
    objects.append( [ Tk.Label( root, text="Hello, world" ), { "fill":"both", "expand":True} ] )

    bf = BlinkingFrame( root, objects )
    Tk.Button( text="Start", command=lambda: bf.start() ).pack()
    Tk.Button( text="Stop", command=lambda: bf.stop() ).pack()

    root.mainloop()
    root.destroy()



"""
Test Excel Tick Labels Formatting
"""
# import pytest
import os
from win32com.client import DispatchEx
import pythoncom
from molass_legacy.KekLib.ExcelCOM import xlCategory, xlValue

# @pytest.fixture(scope="function")
def excel_app():
    pythoncom.CoInitialize()
    excel = DispatchEx("Excel.Application")
    excel.Visible = True
    workbook = excel.Workbooks.Add()
    sheet = workbook.Sheets(1)
    # Add some data
    sheet.Range("A1:A5").Value = [(i,) for i in range(1, 6)]
    sheet.Range("B1:B5").Value = [(i*2,) for i in range(1, 6)]
    # Add a chart (4 = xlLine)
    chart_obj = sheet.Shapes.AddChart2(4, 4, 100, 100, 300, 200)
    chart = chart_obj.Chart
    chart.SetSourceData(sheet.Range("A1:B5"))
    # yield chart
    test_ticklabels_number_format(chart)
    this_dir = os.path.dirname(__file__)
    workbook.SaveAs(os.path.join(this_dir, "test_ticklabels.xlsx"))
    workbook.Close(SaveChanges=False)
    excel.Quit()
    pythoncom.CoUninitialize()

def test_ticklabels_number_format(excel_app):
    chart = excel_app
    # X axis TickLabels (1 = xlCategory)
    x_axis = chart.Axes(xlCategory)
    tick_labels = x_axis.TickLabels
    tick_labels.NumberFormat = "0.0"
    assert tick_labels.NumberFormat == "0.0"

    # Y axis TickLabels (2 = xlValue)
    y_axis = chart.Axes(xlValue)
    y_tick_labels = y_axis.TickLabels
    y_tick_labels.NumberFormat = "0.00E+00"
    assert y_tick_labels.NumberFormat == "0.00E+00"

if __name__ == "__main__":
    excel_app()


"""
    UvSpecCurve.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""


class UvSpecCurve:
    def __init__(self, sd, mpeaks=None):
        D, E, wv, ecurve = sd.get_uv_data_separate_ly()
        self.x = wv
        if mpeaks is None:
            j = ecurve.primary_peak_i
        else:
            xr_ecurve = sd.get_xray_curve()
            j = self.get_matching_peak_j(ecurve, xr_ecurve, mpeaks)
        self.y = D[:,j]

    def get_matching_peak_j(self, ecurve, xr_ecurve, mpeaks):
        j = ecurve.primary_peak_i
        print("ii=", mpeaks.ii)
        print("jj=", mpeaks.jj)
        return j

    def get_x(self):
        return self.x

    def get_y(self):
        return self.y



"""
    SecTheory.SecPDF.py

    Copyright (c) 2022-2024, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.interpolate import UnivariateSpline
from .BasicModels import single_pore_pdf as c
from molass_legacy.KekLib.IntegrateUtils import complex_quadrature_vec

def compute_standard_wCD(N):
    # extracted from .cf2DistFFT.py
    xMin = 0
    xMax = N
    xRange = xMax - xMin
    dt  = 2*np.pi / xRange
    # dt = 1/xRange
    k   = np.arange(N, dtype=complex)     # np.complex is deprecated, or use np.complex128
    w   = (k - N/2 + 0.5) * dt
    A   = xMin
    B   = xMax
    # dx  = (B-A)/N
    c   = (-1)**(A*(N-1)/(B-A))/(B-A)
    # print("A, B, N, dx, c=", A, B, N, dx, c)
    C = c * (-1)**((1-1/N)*k)
    D = (-1)**(-2*(A/(B-A))*k)     # k must be complex, see https://stackoverflow.com/questions/45384602/numpy-runtimewarning-invalid-value-encountered-in-power
    return w, C, D

class FftInvPdf:
    def __init__(self, cf):
        self.cf = cf
        self.N = N = 1024
        self.w, self.C, self.D = compute_standard_wCD(N)

    def __call__(self, t, *params):
        N = self.N
        cft = self.cf(self.w[N//2:], *params)
        cft = np.concatenate([cft[::-1].conj(), cft])
        pdfFFT = np.max([np.zeros(N), (self.C*np.fft.fft(self.D*cft)).real], axis=0)
        spline = UnivariateSpline(np.arange(N), pdfFFT, s=0)
        return spline(t)

class FftInvImpl:
    def __init__(self):
        self.N = N = 1024
        self.w, self.C, self.D = compute_standard_wCD(N)
    
    def get_w(self):
        return self.w

    def compute(self, t, z):
        N = self.N
        z_ = z[N//2:]
        z_ = np.concatenate([z_[::-1].conj(), z_])
        pdfFFT = np.max([np.zeros(N), (self.C*np.fft.fft(self.D*z_)).real], axis=0)
        spline = UnivariateSpline(np.arange(N), pdfFFT, s=0)
        return spline(t)
    
class BidirectFft:
    def __init__(self, x, N=1024):
        self.x = x
        self.N = N 
        self.w, self.C, self.D = compute_standard_wCD(N)
    
    def get_w(self):
        return self.w

    def compute_z(self, y):
        spline = UnivariateSpline(self.x, y, s=0)
        return complex_quadrature_vec(lambda x_: np.exp(1j*self.w*x_)*spline(x_), self.x[0], self.x[-1])[0]

    def compute_y(self, z):
        N = self.N
        z_ = z[N//2:]
        z_ = np.concatenate([z_[::-1].conj(), z_])
        pdfFFT = np.max([np.zeros(N), (self.C*np.fft.fft(self.D*z_)).real], axis=0)
        spline = UnivariateSpline(np.arange(N), pdfFFT, s=0)
        return spline(self.x)


"""
    Kratky.GuinierKratkyPlots.py

    Copyright (c) 2021-2025, SAXS Team, KEK-PF
"""
import numpy as np
from bisect import bisect_right
import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar

def guinier_kratky_plots_impl(ax1, ax2, qv, y, rg, I0, color, interval=None, markersize=3):
    
    # task: consider unifying with Optimizer.GuinierKratkyView.py

    if ax1 is not None:
        from molass.PlotUtils.V1GuinierPlot import guinier_plot
        ax1.set_title("Guinier Plot", fontsize=16)
        guinier_plot(ax1, qv, y, color, interval, markersize)

    if ax2 is not None:
        from molass.PlotUtils.V1KratkyPlot import kratky_plot
        ax2.set_title("Normalized Kratky Plot", fontsize=16)
        kratky_plot(ax2, qv, y, rg, I0, color, markersize)

class GuinierKratkyPlots(Dialog):
    def __init__(self, parent, qv=None, y=None, rg=None, I0=None, range_=None, interval=None, color=None):
        self.qv = qv
        self.y = y
        self.rg = rg
        self.I0 = I0
        self.range_ = range_
        self.interval = interval
        self.color = color
        super().__init__(parent, "GuinierPlot", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):
        cframe = Tk.Frame(body_frame)
        cframe.pack()
        tframe = Tk.Frame(body_frame)
        tframe.pack(fill=Tk.X)

        self.fig, self.axes = plt.subplots(ncols=2, figsize=(14,7))
        self.mpl_canvas = FigureCanvasTkAgg(self.fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)
        self.toolbar = NavigationToolbar(self.mpl_canvas, tframe)
        self.toolbar.update()
        if self.qv is not None:
            self.draw(self.fig, self.axes, self.qv, self.y)
            self.mpl_canvas.draw()

    def draw(self, fig, axes, qv, y):
        range_ = "" if self.range_ is None else "Range [%d, %d]" % self.range_
        title = "Guinier/Kratky Plots for %s" % range_
        interval = self.interval
        if interval is None:
            if self.rg is not None:
                logy = np.log(y)
                interval = self.guess_interval_from_rg(ax1, qv, self.rg, logy)

        fig.suptitle(title, fontsize=20)
        ax1, ax2 = axes

        guinier_kratky_plots_impl(ax1, ax2, qv, y, self.rg, self.I0, self.color, interval)

        fig.tight_layout()
        fig.subplots_adjust(top=0.85)

    def guess_interval_from_rg(self, ax, qq, rg, logy):
        qrg = qq*rg
        i = bisect_right(qrg, 1.3)
        return 0, i



"""
    GuinierTools.CpdDecompIndirectImpl.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize
import ruptures as rpt
import molass_legacy.KekLib.DebugPlot as plt
from GuinierTools.CpdDecompUtils import compute_end_points

VERY_LARGE_VALUE = 1e8
MAX_TAU_RARIO = 1.0

def imporove_decomposition(x, y, model, peaks, end_points):

    new_trs = (end_points[0:-1] + end_points[1:])/2
    min_scale = np.average(peaks[:,0])*0.2
    max_sigma = np.max(peaks[:,2])*1.5

    def near_trs_objective(p):
        cy_list = []
        temp_peaks = p.reshape(peaks.shape)
        for i, params in enumerate(temp_peaks):
            cy = model(x, params)
            cy_list.append(cy)
        ty = np.sum(cy_list, axis=0)
        negetive_penalty = min(0, np.min(temp_peaks[:,0]) - min_scale)**2 * VERY_LARGE_VALUE
        sigma_penalty = max(0, np.max(temp_peaks[:,2]) - max_sigma)**2 * VERY_LARGE_VALUE
        tau_penalty = max(0, np.max(np.abs(temp_peaks[:,3]/temp_peaks[:,2])) - MAX_TAU_RARIO)**2 * VERY_LARGE_VALUE
        return np.log(np.sum((y - ty)**2)) + 1.5*np.log(np.sum((temp_peaks[:,1] - new_trs)**2)) + negetive_penalty + sigma_penalty + tau_penalty

    ret = minimize(near_trs_objective, peaks.flatten(), method='Nelder-Mead')

    return ret.x.reshape(peaks.shape)